import pytest
import sympy

from pharmpy.internals.expr.funcs import PHI
from pharmpy.model import Assignment
from pharmpy.model.external.nonmem.records.code_record import CodeRecord


def _ensure_trailing_newline(buf):
    # FIXME: This should not be necessary
    return buf if buf[-1] == '\n' else buf + '\n'


def S(x):
    return sympy.Symbol(x)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,sym,expression',
    [
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)',
            S('Y'),
            S('THETA(1)') + S('ETA(1)') + S('EPS(1)'),
        ),
        (  # NOTE: github issue #1362
            '$PRED\nY = THETA (1) + ETA(1) + EPS(1)',
            S('Y'),
            S('THETA(1)') + S('ETA(1)') + S('EPS(1)'),
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + ERR(1)',
            S('Y'),
            S('THETA(1)') + S('ETA(1)') + S('EPS(1)'),
        ),
        ('$PRED\nCL = 2', S('CL'), 2),
        ('$PRED\n;FULL LINE COMMENT\n K=-1', S('K'), -1),
        ('$PRED K=-1.5', S('K'), -1.5),
        ('$PRED\nCL = KA', S('CL'), S('KA')),
        ('$PRED\nCL=KA', S('CL'), S('KA')),
        ('$PRED\nG = BASE - LESS', S('G'), S('BASE') - S('LESS')),
        ('$PRED CL = THETA(1) * LEFT', S('CL'), S('THETA(1)') * S('LEFT')),
        ('$PRED D2 = WGT / SRC', S('D2'), S('WGT') / S('SRC')),
        ('$PRED D = W * B + C', S('D'), S('W') * S('B') + S('C')),
        ('$PRED D = W * (B + C)', S('D'), S('W') * (S('B') + S('C'))),
        ('$PRED D = A+B*C+D', S('D'), S('A') + (S('B') * S('C')) + S('D')),
        ('$PRED D = A**2', S('D'), S('A') ** 2),
        ('$PRED D = A - (-2)', S('D'), S('A') + 2),
        ('$PRED D = A - (+2)', S('D'), S('A') - 2),
        ('$PRED D = 2.5', S('D'), 2.5),
        ('$PRED D = 1D-10', S('D'), 1e-10),
        (
            '$PK CL = PLOG(THETA(1))',
            S('CL'),
            sympy.Piecewise(
                (sympy.log(2.8e-103), S('THETA(1)') < 2.8e-103), (sympy.log(S('THETA(1)')), True)
            ),
        ),
        (
            '$PK CL = PLOG10(THETA(1))',
            S('CL'),
            sympy.Piecewise(
                (sympy.log(2.8e-103, 10), S('THETA(1)') < 2.8e-103),
                (sympy.log(S('THETA(1)'), 10), True),
            ),
        ),
        ('$PRED CL = ABS(X)', S('CL'), abs(S('X'))),
        ('$PRED CL = EXP(2)', S('CL'), sympy.exp(2)),
        ('$PRED CL = PEXP(2)', S('CL'), sympy.exp(2)),
        ('$PRED CL = PEXP(101)', S('CL'), sympy.exp(100)),
        ('$PRED CL = exp(2)', S('CL'), sympy.exp(2)),
        ('$PRED cL = eXp(2)', S('CL'), sympy.exp(2)),
        ('$PRED CL = DEXP(ETA(1))', S('CL'), sympy.exp(S('ETA(1)'))),
        (
            '$PRED CL = PDZ(ETA(1))',
            S('CL'),
            sympy.Piecewise((1 / 2.8e-103, abs(S('ETA(1)')) < 2.8e-103), (1 / S('ETA(1)'), True)),
        ),
        (
            '$PRED CL = PZR(ETA(1))',
            S('CL'),
            sympy.Piecewise((2.8e-103, abs(S('ETA(1)')) < 2.8e-103), (S('ETA(1)'), True)),
        ),
        (
            '$PRED CL = PNP(ETA(1))',
            S('CL'),
            sympy.Piecewise((2.8e-103, S('ETA(1)') < 2.8e-103), (S('ETA(1)'), True)),
        ),
        (
            '$PRED CL = PHE(ETA(1))',
            S('CL'),
            sympy.Piecewise((100, S('ETA(1)') > 100), (S('ETA(1)'), True)),
        ),
        (
            '$PRED CL = PNG(ETA(1))',
            S('CL'),
            sympy.Piecewise((0, S('ETA(1)') < 0), (S('ETA(1)'), True)),
        ),
        ('$PRED CL = LOG(V + 1)', S('CL'), sympy.log(S('V') + 1)),
        ('$PRED CL = DLOG(V + 1)', S('CL'), sympy.log(S('V') + 1)),
        ('$PRED CL = LOG10(3.5 + THETA(1))', S('CL'), sympy.log(3.5 + S('THETA(1)'), 10)),
        ('$PRED CL = (SIN(X) + COS(X))', S('CL'), sympy.sin(S('X')) + sympy.cos(S('X'))),
        ('$PRED CL = (DSIN(X) + DCOS(X))', S('CL'), sympy.sin(S('X')) + sympy.cos(S('X'))),
        ('$PRED S22 = ABS(1 + 2 + SIN(X))', S('S22'), sympy.Abs(3 + sympy.sin(S('X')))),
        ('$PRED CL = TAN(X) * EXP(Y)', S('CL'), sympy.tan(S('X')) * sympy.exp(S('Y'))),
        ('$PRED CL = DTAN(X) * DEXP(Y)', S('CL'), sympy.tan(S('X')) * sympy.exp(S('Y'))),
        ('$PRED CL = PTAN(X) * EXP(Y)', S('CL'), sympy.tan(S('X')) * sympy.exp(S('Y'))),
        (
            '$PRED K_ = ATAN(1) - ASIN(X)/ACOS(X)',
            S('K_'),
            sympy.atan(1) - sympy.asin(S('X')) / sympy.acos(S('X')),
        ),
        (
            '$PRED K_ = PATAN(1) - PASIN(X)/PACOS(X)',
            S('K_'),
            sympy.atan(1) - sympy.asin(S('X')) / sympy.acos(S('X')),
        ),
        ('$PRED CL = INT(-2.2)', S('CL'), -2),
        ('$PRED CL = DINT(-2.2)', S('CL'), -2),
        ('$PRED cl = int(-2.2)', S('CL'), -2),
        ('$PRED CL = INT(0.2)', S('CL'), 0),
        ('$PRED CL = SQRT(X)', S('CL'), sympy.sqrt(S('X'))),
        ('$PRED CL = DSQRT(X)', S('CL'), sympy.sqrt(S('X'))),
        (
            '$PRED CL = PSQRT(X)',
            S('CL'),
            sympy.Piecewise((0, S('X') < 0), (sympy.sqrt(S('X')), True)),
        ),
        ('$PRED CL = MOD(1, 2)', S('CL'), sympy.Mod(1, 2)),
        ('$PRED CL = DMOD(1, 2)', S('CL'), sympy.Mod(1, 2)),
        ('$PRED CL = GAMLN(2 + X)   ;COMMENT', S('CL'), sympy.loggamma(S('X') + 2)),
        ('$PRED C02 = PHI(2 + X)', S('C02'), PHI(S('X') + 2)),
        ('$PRED IF (X.EQ.2) CL=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2)), (0, True))),
        ('$PRED IF (X.EQ.2)CL=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2)), (0, True))),
        ('$PRED IF(X.EQ.2)CL=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2)), (0, True))),
        ('$PRED if (x.EQ.2) Cl=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2)), (0, True))),
        (
            '$PRED IF (X.NE.1.5) CL=THETA(1)',
            S('CL'),
            sympy.Piecewise((S('THETA(1)'), sympy.Ne(S('X'), 1.5)), (0, True)),
        ),
        (
            '$PRED IF (X.EQ.2+1) CL=23',
            S('CL'),
            sympy.Piecewise((23, sympy.Eq(S('X'), 3)), (0, True)),
        ),
        (
            '$PRED IF (X < ETA(1)) CL=23',
            S('CL'),
            sympy.Piecewise((23, sympy.Lt(S('X'), S('ETA(1)'))), (0, True)),
        ),
        (
            '$PK IF(AMT.GT.0) BTIME=TIME',
            S('BTIME'),
            sympy.Piecewise((S('TIME'), sympy.Gt(S('AMT'), 0)), (0, True)),
        ),
        (
            '$PRED IF (X.EQ.2.AND.Y.EQ.3) CL=23',
            S('CL'),
            sympy.Piecewise((23, sympy.And(sympy.Eq(S('X'), 2), sympy.Eq(S('Y'), 3))), (0, True)),
        ),
        (
            '$PRED IF (X.EQ.2.OR.Y.EQ.3) CL=23',
            S('CL'),
            sympy.Piecewise((23, sympy.Or(sympy.Eq(S('X'), 2), sympy.Eq(S('Y'), 3))), (0, True)),
        ),
        (
            '$PRED IF (.NOT.X.EQ.2) CL=25',
            S('CL'),
            sympy.Piecewise((25, sympy.Not(sympy.Eq(S('X'), 2))), (0, True)),
        ),
        (
            '$PRED IF (Q.EQ.(R+C)/D) L=0',
            S('L'),
            sympy.Piecewise(
                (0, sympy.Eq(S('Q'), sympy.Mul(sympy.Add(S('R'), S('C')), 1 / S('D')))), (0, True)
            ),
        ),
        (
            '$PRED IF (Q.EQ.R+C/D) L=0',
            S('L'),
            sympy.Piecewise((0, sympy.Eq(S('Q'), S('R') + S('C') / S('D'))), (0, True)),
        ),
        ('$PRED\nA_0(1) = 2', S('A_0(1)'), 2),
        (
            '$PRED\nIF(SPECIES.EQ.4)     IPRG = THETA(49)*EXP(ETA(11))',
            S('IPRG'),
            sympy.Piecewise(
                (S('THETA(49)') * sympy.exp(S('ETA(11)')), sympy.Eq(S('SPECIES'), 4)), (0, True)
            ),
        ),
        ('$PRED\nCL=-KA', S('CL'), -S('KA')),
        ('$PRED\nIFST=-KA', S('IFST'), -S('KA')),
        ('$PRED\nCL=-KA+2', S('CL'), -S('KA') + 2),
        ('$PRED\nCL=-2*X*2', S('CL'), -4 * S('X')),
        ('$PRED\nCL=-(-X)', S('CL'), S('X')),
        (
            '$PRED IF(ROUT.EQ.0.AND.DAYP.EQ.1.AND.PROT.EQ.1088) STRT=10881',
            S('STRT'),
            sympy.Piecewise(
                (
                    10881,
                    sympy.And(
                        sympy.Eq(S('ROUT'), 0), sympy.Eq(S('DAYP'), 1), sympy.Eq(S('PROT'), 1088)
                    ),
                ),
                (0, True),
            ),
        ),
        ('$PRED X=A+B+C', S('X'), S('A') + S('B') + S('C')),
        ('$PRED X=A+B-C', S('X'), S('A') + S('B') - S('C')),
        ('$PRED X=A+B*C', S('X'), S('A') + S('B') * S('C')),
        ('$PRED X=A*B+C', S('X'), S('A') * S('B') + S('C')),
        ('$PRED\n"VERBATIM STUFF\nK=1', S('K'), 1),
        ('$PRED\n"VERBATIM STUFF\n"ON TWO LINES\nK=1', S('K'), 1),
        ('$PRED (CALLFL=0)\n\nK=1', S('K'), 1),
        ('$ERROR\nCL = 2', S('CL'), 2),
        ('$ERROR (ONLY OBSERVATION) \nCL = 2', S('CL'), 2),
        ('$ERROR\n (ONLY OBSERVATION) \nCL = 2', S('CL'), 2),
        ('$ERROR\n(ONLY OBSERVATION)\nCL = 2', S('CL'), 2),
        ('$PRED\nCL = 2\nEXIT\n', S('CL'), 2),
        ('$PRED\nCL = 2\nEXIT \n', S('CL'), 2),
        ('$PRED\nCL = 2\nEXIT 1 \n', S('CL'), 2),
        ('$PRED\nCL = 2\nEXIT 1 23 \n', S('CL'), 2),
        ('$PRED\nCL = 2\nCALL RANDOM (2, R)\n', S('CL'), 2),
        ('$PRED\nCL = 2\n  RETURN  \n', S('CL'), 2),
        ('$PRED\nCOM(1) = 0\n', S('COM(1)'), 0),
        ('$PRED\nANYTHING(1) = 0\n', S('ANYTHING(1)'), 0),
        ('$PRED\nGROK(1,3) = 0\n', S('GROK(1,3)'), 0),
        ('$PRED\n X = &  \n  23\n', S('X'), 23),
        ('$PRED\nX=2\nDO WHILE (X < 23)\n CL = 2\nEND DO\n', S('X'), 2),
        (  # NOTE: Regression when we switched to LALR(1)
            '$PK\nALAG1 = EXP(PHI+KPLAG)/(1+EXP(PHI+KPLAG))',
            S('ALAG1'),
            sympy.exp(S('PHI') + S('KPLAG'))
            / (sympy.Integer(1) + sympy.exp(S('PHI') + S('KPLAG'))),
        ),
        (
            '$PK\nV=TVV*EXP(ETA2)',
            S('V'),
            S('TVV') * sympy.exp(S('ETA2')),
        ),
        (
            '$PK\nV=THETA\t \x00\t \x00\t \x00(1)',
            S('V'),
            S('THETA(1)'),
        ),
        (
            '$PK\nV=THETA \t & \t\t\x00\r\n & \t\t\x00\n  (1)',
            S('V'),
            S('THETA(1)'),
        ),
        ('$PK\nX(Z)=0', S('X(Z)'), 0),
        ('$PK\nX(  Z \x00)=0', S('X(Z)'), 0),
        ('$PK\n X \t \t \x00\x00\t \x00(\x00Z)=0', S('X(Z)'), 0),
        ('$PK\n X \t \t \x00& \t\t\x00\n \t \x00( -1 , 2 )=0', S('X(-1,2)'), 0),
    ],
)
def test_single_assignments(parser, buf, sym, expression):
    buf = _ensure_trailing_newline(buf)
    rec = parser.parse(buf).records[0]
    assert len(rec.statements) == 1
    assert rec.statements[0].symbol == sym
    assert rec.statements[0].expression == expression


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,symb_expr_arr',
    [
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23\nZ = 9\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (FA1.EQ.0) IOV_1 = ETA(3)\nIF (FA1.EQ.1) IOV_1 = ETA(4)',
            [
                (S('IOV_1'), sympy.Piecewise((S('ETA(3)'), sympy.Eq(S('FA1'), 0)), (0, True))),
                (
                    S('IOV_1'),
                    sympy.Piecewise((S('ETA(4)'), sympy.Eq(S('FA1'), 1)), (S('IOV_1'), True)),
                ),
            ],
        ),
        (
            '$PRED IF (B0.LT.3) THEN\nCL = THETA(1)\nELSE\nCL = 23\nENDIF',
            [(S('CL'), sympy.Piecewise((S('THETA(1)'), S('B0') < 3), (23, True)))],
        ),
        (
            '$PRED IF (B0.LT.3) THEN\nCL = THETA(1)\nKA = THETA(2)\n  ELSE  \nCL = 23\nKA=82\n'
            'END IF',
            [
                (S('CL'), sympy.Piecewise((S('THETA(1)'), S('B0') < 3), (23, True))),
                (S('KA'), sympy.Piecewise((S('THETA(2)'), S('B0') < 3), (82, True))),
            ],
        ),
        (
            '$PRED    IF (A>=.5) THEN   \n  VAR=1+2 \nELSE IF (B.EQ.23) THEN ; C\nVAR=9.25\n'
            'END IF  \n',
            [
                (
                    S('VAR'),
                    sympy.Piecewise((3, S('A') >= 0.5), (9.25, sympy.Eq(S('B'), 23)), (0, True)),
                )
            ],
        ),
        (
            '$PRED    if (a>=0.5) then   \n  var=1+2 \nELSE if (b.eq.23) then ; C\nVAR=9.25\n'
            'END IF  \n',
            [
                (
                    S('VAR'),
                    sympy.Piecewise((3, S('A') >= 0.5), (9.25, sympy.Eq(S('B'), 23)), (0, True)),
                )
            ],
        ),
        (
            '$PRED   IF (A>=0.5) THEN   \n  VAR1=1+2 \nELSE IF  (B.EQ.23)  THEN \nVAR2=9.25\n'
            'END IF  \n',
            [
                (S('VAR1'), sympy.Piecewise((3, S('A') >= 0.5), (0, True))),
                (S('VAR2'), sympy.Piecewise((9.25, sympy.Eq(S('B'), 23)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN    ;   anything  \n  Y = 23\nZ = 9\nEND IF;AFTER',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (NEWIND.NE.2.OR.EVID.GE.3) THEN ; begin\nTNXD=TIME ; TIME\nENDIF',
            [
                (
                    S('TNXD'),
                    sympy.Piecewise(
                        (S('TIME'), sympy.Or(sympy.Ne(S('NEWIND'), 2), sympy.Ge(S('EVID'), 3))),
                        (0, True),
                    ),
                )
            ],
        ),
        (
            '$PRED\nIF (NEWIND.NE.2.OR.EVID.GE.3.) THEN ; begin\nTNXD=TIME ; TIME\nENDIF',
            [
                (
                    S('TNXD'),
                    sympy.Piecewise(
                        (S('TIME'), sympy.Or(sympy.Ne(S('NEWIND'), 2), sympy.Ge(S('EVID'), 3.0))),
                        (0, True),
                    ),
                )
            ],
        ),
        (
            '$PRED\nIF (NEWIND.NE.2..OR.EVID.GE.3) THEN ; begin\nTNXD=TIME ; TIME\nENDIF',
            [
                (
                    S('TNXD'),
                    sympy.Piecewise(
                        (S('TIME'), sympy.Or(sympy.Ne(S('NEWIND'), 2.0), sympy.Ge(S('EVID'), 3))),
                        (0, True),
                    ),
                )
            ],
        ),
        (
            '$PRED IF (B0.LT.3) THEN\nCL = THETA(1)\nELSE;A close comment\nCL = 23\nENDIF',
            [(S('CL'), sympy.Piecewise((S('THETA(1)'), S('B0') < 3), (23, True)))],
        ),
        (
            '$PRED\nIF(MIXNUM.EQ.3) THEN\n    TVCL=THETA(1)       ; CL in population 1\nELSE\n    '
            'TVCL=THETA(2)       ; CL in population 2\nENDIF\n',
            [
                (
                    S('TVCL'),
                    sympy.Piecewise(
                        (S('THETA(1)'), sympy.Eq(S('MIXNUM'), 3)), (S('THETA(2)'), True)
                    ),
                )
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23.\nZ = 9\nELSE\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23.0, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23\nZ = 9.\nELSE IF (X.EQ.23) THEN\nELSE\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9.0, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23\nZ = 9\nELSE IF (X.EQ.44) THEN\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nELSE\nY = 23\nZ = 9\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Ne(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Ne(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23\nZ = 9\nELSE\n CALL RANDOM(1,R)\nEND IF',
            [
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (0, True))),
                (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0)), (0, True))),
            ],
        ),
        (
            '$PRED\nIF (X.EQ.0) THEN\nY = 23\nELSE\nY = 7\nZ = 9\nEND IF',
            [
                (S('Z'), 9),
                (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)), (7, True))),
            ],
        ),
    ],
)
def test_block_if(parser, buf, symb_expr_arr):
    buf = _ensure_trailing_newline(buf)
    rec = parser.parse(buf).records[0]
    assert len(rec.statements) == len(symb_expr_arr)
    for statement, (symb, expr) in zip(rec.statements, symb_expr_arr):
        assert statement.symbol == symb
        assert statement.expression == expr


@pytest.mark.parametrize(
    'buf',
    (
        '$PK IF (CL.EQ.0) EXIT 1',
        '$PK IF (CL.EQ.0) EXIT 1 24',
    ),
)
def test_exit(parser, buf):
    buf = _ensure_trailing_newline(buf)
    rec = parser.parse(buf).records[0]
    assert len(rec.statements) == 0


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf',
    [
        '$PRED\nY = THETA(1) + ETA(1) + EPS(1)',
        '$PRED IF (X.EQ.2) CL=23',
        '$PRED IF (X.NE.1.5) CL=THETA(1)',
        '$PRED IF (X.EQ.2+1) CL=23',
        '$PRED IF (X.EQ.2.AND.Y.EQ.3) CL=23',
        '$PRED IF (X.EQ.2.OR.Y.EQ.3) CL=23',
        '$PRED IF (.NOT.X.EQ.2) CL=25',
    ],
)
def test_grammar_repeats(parser, buf):  # Tests that there are no repeats due to parsing grammar
    buf = _ensure_trailing_newline(buf)
    rec = parser.parse(buf).records[0]
    tree_walk_gen = rec.root.tree_walk()
    parent = next(tree_walk_gen)

    for child in tree_walk_gen:
        assert parent != child
        parent = child


def test_pheno(parser):
    code = """$PK

IF(AMT.GT.0) BTIME=TIME
TAD=TIME-BTIME
      TVCL=THETA(1)*WGT
      TVV=THETA(2)*WGT
IF(APGR.LT.5) TVV=TVV*(1+THETA(3))
      CL=TVCL*EXP(ETA(1))
      V=TVV*EXP(ETA(2))
      S1=V
"""
    rec = parser.parse(code).records[0]
    assert len(rec.statements) == 8
    assert rec.statements[0].symbol == S('BTIME')
    assert rec.statements[1].symbol == S('TAD')
    assert rec.statements[2].symbol == S('TVCL')
    assert rec.statements[3].symbol == S('TVV')
    assert rec.statements[4].symbol == S('TVV')
    assert rec.statements[5].symbol == S('CL')
    assert rec.statements[6].symbol == S('V')
    assert rec.statements[7].symbol == S('S1')
    assert rec.statements[0].expression == sympy.Piecewise(
        (S('TIME'), sympy.Gt(S('AMT'), 0)), (0, True)
    )
    assert rec.statements[1].expression == S('TIME') - S('BTIME')
    assert rec.statements[2].expression == S('THETA(1)') * S('WGT')
    assert rec.statements[3].expression == S('THETA(2)') * S('WGT')
    assert rec.statements[4].expression == sympy.Piecewise(
        (S('TVV') * (1 + S('THETA(3)')), sympy.Lt(S('APGR'), 5)), (S('TVV'), True)
    )
    assert rec.statements[5].expression == S('TVCL') * sympy.exp(S('ETA(1)'))
    assert rec.statements[6].expression == S('TVV') * sympy.exp(S('ETA(2)'))
    assert rec.statements[7].expression == S('V')
    symbol_names = {s.name for s in rec.statements.free_symbols}
    assert symbol_names == {
        'AMT',
        'BTIME',
        'TIME',
        'TAD',
        'TVCL',
        'THETA(1)',
        'WGT',
        'TVV',
        'THETA(2)',
        'APGR',
        'THETA(3)',
        'CL',
        'ETA(1)',
        'V',
        'ETA(2)',
        'S1',
    }


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,buf_new',
    [
        ('$PRED\nY = THETA(1) + ETA(1) + EPS(1)', '$PRED\nY = THETA(1) + ETA(1) + EPS(1)'),
        ('$PRED\nY = THETA(1) + ETA(1) + EPS(1) ;comment', '$PRED\nY = THETA(1) + ETA(1) + EPS(1)'),
        ('$PRED\nY = THETA(1) + ETA(1) + EPS(1)', '$PRED\nY=THETA(1)+ETA(1)+EPS(1)'),
    ],
)
def test_statements_setter_identical(parser, buf_original, buf_new):
    buf_original = _ensure_trailing_newline(buf_original)
    buf_new = _ensure_trailing_newline(buf_new)
    rec_original = parser.parse(buf_original).records[0]
    rec_new = parser.parse(buf_new).records[0]

    assert rec_original.statements == rec_new.statements


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,buf_new',
    [
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
        ),
        ('$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2', '$PRED\nCL = 2'),
        ('$PRED\nY = A + B\nX = C - D\nZ = E * F', '$PRED\nY = A + B\nZ = E * F'),
    ],
)
def test_statements_setter_remove(parser, buf_original, buf_new):
    buf_original = _ensure_trailing_newline(buf_original)
    buf_new = _ensure_trailing_newline(buf_new)
    rec_original = parser.parse(buf_original).records[0]
    rec_new = parser.parse(buf_new).records[0]

    newrec = rec_original.update_statements(rec_new.statements)

    assert newrec.statements == rec_new.statements
    assert tuple(newrec.root.subtrees('statement')) == tuple(rec_new.root.subtrees('statement'))


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,buf_new',
    [
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nS1 = V\nCL = 2\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nS1 = V\n',
            '$PRED\nCL = 2\nY = THETA(1) + ETA(1) + EPS(1)\nS1 = V\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2\n',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nS1 = V\nCL = 2\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2\n',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nS1 = V\nTVCL = WGT + 2\nCL = 2\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nIF (AMT.GT.0) BTIME = TIME\n',
        ),
        (
            '$PRED\nBTIME = 2\nIF (AMT.GT.0) BTIME = TIME\n',
            '$PRED\nCL = 2\nBTIME = 2\nIF (AMT.GT.0) BTIME = TIME\n',
        ),
    ],
)
def test_statements_setter_add(parser, buf_original, buf_new):
    rec_original = parser.parse(buf_original).records[0]
    rec_new = parser.parse(buf_new).records[0]

    newrec = rec_original.update_statements(rec_new.statements)
    assert newrec.statements == rec_new.statements
    assert tuple(newrec.root.subtrees('statement')) == tuple(rec_new.root.subtrees('statement'))


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,buf_new',
    [
        ('$PRED\nY = A + B\n', '$PRED\nZ = E*F\n'),
        ('$PRED\nY = A + B\nX = C - D\nZ = E * F\n', '$PRED\nY = A + B\nX = C\nZ = E * F\n'),
        ('$PRED\nY = A + B\nX = C - D\nZ = E * F\n', '$PRED\nY = A + B\nX = C - D\nZ = E\n'),
        ('$PRED\nY = A + B\nX = C - D\nZ = E * F\n', '$PRED\nY = A + B\nX = C\nZ = E * F\n'),
    ],
)
def test_statements_setter_change(parser, buf_original, buf_new):
    rec_original = parser.parse(buf_original).records[0]
    rec_new = parser.parse(buf_new).records[0]
    newrec = rec_original.update_statements(rec_new.statements)

    assert newrec.statements == rec_new.statements
    assert tuple(newrec.root.subtrees('statement')) == tuple(rec_new.root.subtrees('statement'))
    assert str(newrec) == buf_new


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,sym,expression,buf_new',
    [
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            S('CL'),
            2,
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 2\n',
        ),
        (
            '$PRED\n"FIRST\n"!Fortran code goes here\n',
            S('V'),
            -S('CL'),
            '$PRED\nV = -CL\n"FIRST\n"!Fortran code goes here\n',
        ),
    ],
)
def test_statements_setter_add_from_sympy(parser, buf_original, sym, expression, buf_new):
    rec_original = parser.parse(buf_original).records[0]

    assignment = Assignment(sym, expression)
    statements = rec_original.statements + [assignment]
    newrec = rec_original.update_statements(statements)

    assert str(newrec) == buf_new


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf_original,assignment,buf_expected',
    [
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            Assignment(S('Z'), S('THETA(2)')),
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nZ = THETA(2)\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 1.3\n',
            Assignment(S('Z'), S('THETA(2)')),
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\nCL = 1.3\nZ = THETA(2)\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            Assignment(S('YWGT'), sympy.Piecewise((1, sympy.Eq(S('WGT'), S('NaN'))))),
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n' 'IF (NaN.EQ.WGT) YWGT = 1\n',
        ),
        (
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n',
            Assignment(S('Z'), PHI(S('A') + S('B'))),
            '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n' 'Z = PHI(A + B)\n',
        ),
    ],
)
def test_update(parser, buf_original, assignment, buf_expected):
    rec_original = parser.parse(buf_original).records[0]

    statements = rec_original.statements + assignment
    newrec = rec_original.update_statements(statements)

    assert str(newrec) == buf_expected


def test_nested_block_if(parser):
    code = (
        '\nIF (X.EQ.23) THEN\nIF (Y.EQ.0) THEN\nCL=1\nELSE\nCL=2\nENDIF\n' 'CL=5\nELSE\nCL=6\nENDIF'
    )
    code = _ensure_trailing_newline(code)
    rec = parser.parse('$PRED' + code).records[0]

    s = rec.statements
    rec = rec.update_statements(s)
    assert str(rec.root) == code


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'symbol, expression, buf_expected',
    [
        (
            S('BTIME'),
            sympy.Piecewise((S('TIME'), sympy.Gt(S('AMT'), 0))),
            'IF (AMT.GT.0) BTIME = TIME\n',
        ),
        (S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2))), 'IF (X.EQ.2) CL = 23\n'),
        (
            S('CLWGT'),
            sympy.Piecewise((23, sympy.Eq(S('X'), 1)), (0, sympy.Eq(S('X'), 0))),
            'IF (X.EQ.1) CLWGT = 23\nIF (X.EQ.0) CLWGT = 0\n',
        ),
        (
            S('X'),
            sympy.Piecewise((0, sympy.And(sympy.Ne(S('Y'), 1), sympy.Ne(S('K'), 2)))),
            'IF (K.NE.2.AND.Y.NE.1) X = 0\n',
        ),
        (
            S('X'),
            sympy.Piecewise((0, sympy.Ne(S('Y'), 1)), (1, True)),
            'IF (Y.NE.1) THEN\n    X = 0\nELSE\n    X = 1\nEND IF\n',
        ),
        (
            S('X'),
            sympy.Piecewise((0, sympy.Ne(S('Y'), 1)), (2, sympy.Eq(S('Y'), 15)), (1, True)),
            'IF (Y.NE.1) THEN\n    X = 0\nELSE IF (Y.EQ.15) THEN\n    X = 2\nELSE\n    X = 1\nEND IF\n',  # noqa: E501
        ),
    ],
)
def test_translate_sympy_piecewise(parser, symbol, expression, buf_expected):
    buf_original = '$PRED\nY = THETA(1) + ETA(1) + EPS(1)\n'
    rec = parser.parse(buf_original).records[0]
    assert isinstance(rec, CodeRecord)
    s = Assignment(symbol, expression)
    statements = rec.statements + s
    newrec = rec.update_statements(statements)
    assert str(newrec) == buf_original + buf_expected


def test_empty_record(parser):
    rec = parser.parse('$PRED\n').records[0]
    assert len(rec.statements) == 0


def test_set_block_if(parser):
    rec = parser.parse('$PRED\n').records[0]
    z = S('z')
    s = Assignment(sympy.Symbol('X'), sympy.Piecewise((23, z < 12), (5, True)))
    statements = rec.statements + s
    rec = rec.update_statements(statements)
    correct = """$PRED
IF (Z.LT.12) THEN
    X = 23
ELSE
    X = 5
END IF
"""
    assert str(rec) == correct



import re
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.deps.rich import box as rich_box
from pharmpy.deps.rich import console as rich_console
from pharmpy.deps.rich import table as rich_table
from pharmpy.model import ColumnInfo, CompartmentalSystem, DataInfo, DatasetError, Model
from pharmpy.model.model import update_datainfo

from .iterators import resample_data


def get_ids(model: Model):
    """Retrieve a list of all subject ids of the dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    list
        All subject ids

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, get_ids
    >>> model = load_example_model("pheno")
    >>> get_ids(model)      # doctest: +ELLIPSIS
    [1, 2, 3, ..., 57, 58, 59]
    """
    idcol = model.datainfo.id_column.name
    ids = list(model.dataset[idcol].unique())
    return ids


def get_number_of_individuals(model: Model):
    """Retrieve the number of individuals in the model dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of individuals in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_individuals, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_individuals(model)
    59

    Notes
    -----
    For NONMEM models this is the number of individuals of the active dataset, i.e. after filtering
    of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_observations : Get the number of observations in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    return len(get_ids(model))


def get_number_of_observations(model: Model):
    """Retrieve the total number of observations in the model dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of observations in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_observations, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_observations(model)
    155

    Notes
    -----
    For NONMEM models this is the number of observations of the active dataset, i.e. after filtering
    of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_individuals : Get the number of individuals in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    return len(get_observations(model))


def get_number_of_observations_per_individual(model: Model):
    """Number of observations for each individual

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        Number of observations in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_observations_per_individual, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_observations_per_individual(model)
    ID
    1     2
    2     3
    3     3
    4     3
    5     3
    6     3
    7     3
    8     3
    9     4
    10    3
    11    1
    12    3
    13    2
    14    4
    15    2
    16    3
    17    3
    18    4
    19    3
    20    3
    21    3
    22    2
    23    3
    24    3
    25    6
    26    2
    27    2
    28    1
    29    1
    30    2
    31    1
    32    3
    33    2
    34    2
    35    2
    36    3
    37    2
    38    4
    39    3
    40    2
    41    3
    42    2
    43    1
    44    3
    45    3
    46    1
    47    1
    48    5
    49    3
    50    4
    51    3
    52    3
    53    2
    54    4
    55    1
    56    1
    57    2
    58    3
    59    3
    Name: observation_count, dtype: int64

    Notes
    -----
    For NONMEM models this is the individuals and number of observations of the active dataset, i.e.
    after filtering of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_individuals : Get the number of individuals in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    ser = get_observations(model).groupby(model.datainfo.id_column.name).count()
    ser.name = "observation_count"
    return ser


def get_observations(model: Model, keep_index: bool = False) -> pd.Series:
    """Get observations from dataset

    Parameters
    ----------
    model : Model
        Pharmpy model
    keep_index : bool
        Set to True if the original index should be kept.
        Otherwise a new index using ID and idv will be created.

    Returns
    -------
    pd.Series
        Observations indexed over ID and TIME

    Examples
    --------
    >>> from pharmpy.modeling import get_observations, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_observations(model)
        ID  TIME
    1   2.0      17.3
        112.5    31.0
    2   2.0       9.7
        63.5     24.6
        135.5    33.0
                 ...
    58  47.5     27.9
        131.8    31.0
    59  1.8      22.6
        73.8     34.3
        146.8    40.2
    Name: DV, Length: 155, dtype: float64

    See also
    --------
    get_number_of_observations : get the number of observations
    get_number_of_observations_per_individual : get the number of observations per individual
    """
    try:
        label = model.datainfo.typeix['mdv'][0].name
    except IndexError:
        try:
            label = model.datainfo.typeix['event'][0].name
        except IndexError:
            try:
                label = model.datainfo.typeix['dose'][0].name
            except IndexError:
                label = None  # All data records are observations

    idcol = model.datainfo.id_column.name
    idvcol = model.datainfo.idv_column.name
    dvcol = model.datainfo.dv_column.name

    if label:
        df = model.dataset.query(f'{label} == 0')
        if df.empty:
            df = model.dataset.astype({label: 'float'})
            df = df.query(f'{label} == 0')
    else:
        df = model.dataset.copy()

    if not keep_index:
        df = df[[idcol, idvcol, dvcol]]
        try:
            # FIXME: This shouldn't be needed
            df = df.astype({idvcol: np.float64})
        except ValueError:
            # TIME could not be converted to float (e.g. 10:15)
            pass
        df.set_index([idcol, idvcol], inplace=True)
        df = df.squeeze()
    else:
        df = df[dvcol]
    return df


def get_baselines(model: Model):
    """Baselines for each subject.

    Baseline is taken to be the first row even if that has a missing value.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.DataFrame
        Dataset with the baselines

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_baselines
    >>> model = load_example_model("pheno")
    >>> get_baselines(model)
         TIME  AMT  WGT  APGR   DV  FA1  FA2
    ID
    1    0.0  25.0  1.4   7.0  0.0  1.0  1.0
    2    0.0  15.0  1.5   9.0  0.0  1.0  1.0
    3    0.0  30.0  1.5   6.0  0.0  1.0  1.0
    4    0.0  18.6  0.9   6.0  0.0  1.0  1.0
    5    0.0  27.0  1.4   7.0  0.0  1.0  1.0
    6    0.0  24.0  1.2   5.0  0.0  1.0  1.0
    7    0.0  19.0  1.0   5.0  0.0  1.0  1.0
    8    0.0  24.0  1.2   7.0  0.0  1.0  1.0
    9    0.0  27.0  1.4   8.0  0.0  1.0  1.0
    10   0.0  27.0  1.4   7.0  0.0  1.0  1.0
    11   0.0  24.0  1.2   7.0  0.0  1.0  1.0
    12   0.0  26.0  1.3   6.0  0.0  1.0  1.0
    13   0.0  11.0  1.1   6.0  0.0  1.0  1.0
    14   0.0  22.0  1.1   7.0  0.0  1.0  1.0
    15   0.0  26.0  1.3   7.0  0.0  1.0  1.0
    16   0.0  12.0  1.2   9.0  0.0  1.0  1.0
    17   0.0  22.0  1.1   5.0  0.0  1.0  1.0
    18   0.0  20.0  1.0   5.0  0.0  1.0  1.0
    19   0.0  10.0  1.0   1.0  0.0  1.0  1.0
    20   0.0  24.0  1.2   6.0  0.0  1.0  1.0
    21   0.0  17.5  1.8   7.0  0.0  1.0  1.0
    22   0.0  15.0  1.5   8.0  0.0  1.0  1.0
    23   0.0  60.0  3.1   3.0  0.0  1.0  1.0
    24   0.0  63.0  3.2   2.0  0.0  1.0  1.0
    25   0.0  15.0  0.7   1.0  0.0  1.0  1.0
    26   0.0  70.0  3.5   9.0  0.0  1.0  1.0
    27   0.0  35.0  1.9   5.0  0.0  1.0  1.0
    28   0.0  60.0  3.2   9.0  0.0  1.0  1.0
    29   0.0  20.0  1.0   7.0  0.0  1.0  1.0
    30   0.0  18.0  1.8   8.0  0.0  1.0  1.0
    31   0.0  30.0  1.4   8.0  0.0  1.0  1.0
    32   0.0  70.0  3.6   9.0  0.0  1.0  1.0
    33   0.0  17.0  1.7   8.0  0.0  1.0  1.0
    34   0.0  34.0  1.7   4.0  0.0  1.0  1.0
    35   0.0  25.0  2.5   5.0  0.0  1.0  1.0
    36   0.0  30.0  1.5   5.0  0.0  1.0  1.0
    37   0.0  24.0  1.2   9.0  0.0  1.0  1.0
    38   0.0  26.0  1.3   8.0  0.0  1.0  1.0
    39   0.0  56.0  1.9  10.0  0.0  1.0  1.0
    40   0.0  19.0  1.1   3.0  0.0  1.0  1.0
    41   0.0  34.0  1.7   7.0  0.0  1.0  1.0
    42   0.0  28.0  2.8   9.0  0.0  1.0  1.0
    43   0.0  18.0  0.9   1.0  0.0  1.0  1.0
    44   0.0  14.0  1.4   7.0  0.0  1.0  1.0
    45   0.0  16.0  0.8   2.0  0.0  1.0  1.0
    46   0.0  11.0  1.1   8.0  0.0  1.0  1.0
    47   0.0  40.0  2.6   9.0  0.0  1.0  1.0
    48   0.0  14.0  0.7   8.0  0.0  1.0  1.0
    49   0.0  26.0  1.3   8.0  0.0  1.0  1.0
    50   0.0  20.0  1.1   6.0  0.0  1.0  1.0
    51   0.0  18.0  0.9   9.0  0.0  1.0  1.0
    52   0.0   9.5  0.9   7.0  0.0  1.0  1.0
    53   0.0  17.0  1.7   8.0  0.0  1.0  1.0
    54   0.0  18.0  1.8   8.0  0.0  1.0  1.0
    55   0.0  25.0  1.1   4.0  0.0  1.0  1.0
    56   0.0  12.0  0.6   4.0  0.0  1.0  1.0
    57   0.0  20.0  2.1   6.0  0.0  1.0  1.0
    58   0.0  14.0  1.4   8.0  0.0  1.0  1.0
    59   0.0  22.8  1.1   6.0  0.0  1.0  1.0
    """
    idlab = model.datainfo.id_column.name
    baselines = model.dataset.groupby(idlab).nth(0).set_index(idlab)
    return baselines


def set_covariates(model: Model, covariates: List[str]):
    """Set columns in the dataset to be covariates in the datainfo

    Parameters
    ----------
    model : Model
        Pharmpy model
    covariates : list
        List of column names

    Returns
    -------
    Model
        Pharmpy model object
    """
    di = model.datainfo
    newcols = []
    for col in di:
        if col.name in covariates:
            newcol = col.replace(type='covariate')
            newcols.append(newcol)
        else:
            newcols.append(col)
    model = model.replace(datainfo=di.replace(columns=newcols))
    return model.update_source()


def set_dvid(model: Model, name: str):
    """Set a column to act as DVID. Replace DVID if one is already set.

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of DVID column

    Returns
    -------
    Model
        Pharmpy model object
    """
    di = model.datainfo
    col = di[name]
    if col.type == 'dvid':
        return model

    try:
        curdvid = di.typeix['dvid'][0]
    except IndexError:
        pass
    else:
        curdvid = curdvid.replace(type='unknown')
        di = di.set_column(curdvid)

    col = col.replace(
        type='dvid',
        unit=1,
        scale='nominal',
        continuous=False,
        drop=False,
        descriptor='observation identifier',
    )
    df = model.dataset
    if not col.is_integer():
        ser = df[name]
        converted = pd.to_numeric(ser, downcast='integer')
        if not pd.api.types.is_integer_dtype(converted):
            raise ValueError(
                f"Could not use column {name} as DVID because it contains non-integral values"
            )
        df = df.assign(**{name: converted})
        col = col.replace(datatype=ColumnInfo.convert_pd_dtype_to_datatype(converted.dtype))
        new_dataset = True
    else:
        new_dataset = False

    col = col.replace(categories=sorted(df[name].unique()))

    di = di.set_column(col)

    if new_dataset:
        model = model.replace(datainfo=di, dataset=df)
    else:
        model = model.replace(datainfo=di)
    return model.update_source()


def get_covariate_baselines(model: Model):
    """Return a dataframe with baselines of all covariates for each id.

    Baseline is taken to be the first row even if that has a missing value.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.DataFrame
        covariate baselines

    See also
    --------
    get_baselines : baselines for all data columns

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, get_covariate_baselines, set_covariates
    >>> model = load_example_model("pheno")
    >>> model = set_covariates(model, ["WGT", "APGR"])
    >>> get_covariate_baselines(model)
        WGT  APGR
    ID
    1   1.4   7.0
    2   1.5   9.0
    3   1.5   6.0
    4   0.9   6.0
    5   1.4   7.0
    6   1.2   5.0
    7   1.0   5.0
    8   1.2   7.0
    9   1.4   8.0
    10  1.4   7.0
    11  1.2   7.0
    12  1.3   6.0
    13  1.1   6.0
    14  1.1   7.0
    15  1.3   7.0
    16  1.2   9.0
    17  1.1   5.0
    18  1.0   5.0
    19  1.0   1.0
    20  1.2   6.0
    21  1.8   7.0
    22  1.5   8.0
    23  3.1   3.0
    24  3.2   2.0
    25  0.7   1.0
    26  3.5   9.0
    27  1.9   5.0
    28  3.2   9.0
    29  1.0   7.0
    30  1.8   8.0
    31  1.4   8.0
    32  3.6   9.0
    33  1.7   8.0
    34  1.7   4.0
    35  2.5   5.0
    36  1.5   5.0
    37  1.2   9.0
    38  1.3   8.0
    39  1.9  10.0
    40  1.1   3.0
    41  1.7   7.0
    42  2.8   9.0
    43  0.9   1.0
    44  1.4   7.0
    45  0.8   2.0
    46  1.1   8.0
    47  2.6   9.0
    48  0.7   8.0
    49  1.3   8.0
    50  1.1   6.0
    51  0.9   9.0
    52  0.9   7.0
    53  1.7   8.0
    54  1.8   8.0
    55  1.1   4.0
    56  0.6   4.0
    57  2.1   6.0
    58  1.4   8.0
    59  1.1   6.0
    """
    covariates = model.datainfo.typeix['covariate'].names
    idlab = model.datainfo.id_column.name
    df = model.dataset[covariates + [idlab]]
    df = df.set_index(idlab)
    return df.groupby(idlab).nth(0)


def list_time_varying_covariates(model: Model):
    """Return a list of names of all time varying covariates

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    list
        Names of all time varying covariates

    See also
    --------
    get_covariate_baselines : get baselines for all covariates

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, list_time_varying_covariates
    >>> model = load_example_model("pheno")
    >>> list_time_varying_covariates(model)
    []

    """
    cov_labels = model.datainfo.typeix['covariate'].names
    if len(cov_labels) == 0:
        return []
    else:
        time_var = (
            model.dataset.groupby(by=model.datainfo.id_column.name)[cov_labels]
            .nunique()
            .gt(1)
            .any()
        )
        return list(time_var.index[time_var])


def get_doses(model: Model):
    """Get a series of all doses

    Indexed with ID and TIME

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        doses

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, get_doses
    >>> model = load_example_model("pheno")
    >>> get_doses(model)
    ID  TIME
    1   0.0      25.0
        12.5      3.5
        24.5      3.5
        37.0      3.5
        48.0      3.5
                 ...
    59  96.0      3.0
        108.3     3.0
        120.5     3.0
        132.3     3.0
        144.8     3.0
    Name: AMT, Length: 589, dtype: float64

    """
    try:
        label = model.datainfo.typeix['dose'][0].name
    except IndexError:
        raise DatasetError('Could not identify dosing rows in dataset')

    idcol = model.datainfo.id_column.name
    idvcol = model.datainfo.idv_column.name
    df = model.dataset.query(f'{label} != 0')
    df = df[[idcol, idvcol, label]]
    try:
        # FIXME: This shouldn't be needed
        df = df.astype({idvcol: np.float64})
    except ValueError:
        # TIME could not be converted to float (e.g. 10:15)
        pass
    df.set_index([idcol, idvcol], inplace=True)
    return df.squeeze()


def expand_additional_doses(model: Model, flag: bool = False):
    """Expand additional doses into separate dose records

    Parameters
    ----------
    model : Model
        Pharmpy model object
    flag : bool
        True to add a boolean EXPANDED column to mark added records. In this case all
        columns in the original dataset will be kept. Care needs to be taken to handle
        the new dataset.

    Returns
    -------
    Model
        Pharmpy model object
    """
    try:
        addl = model.datainfo.typeix['additional'][0].name
        ii = model.datainfo.typeix['ii'][0].name
    except IndexError:
        return model
    idv = model.datainfo.idv_column.name
    idcol = model.datainfo.id_column.name

    df = model.dataset.copy()

    try:
        event = model.datainfo.typeix['event'][0].name
    except IndexError:
        df['_RESETGROUP'] = 1.0
    else:
        df['_FLAG'] = df[event] >= 3
        df['_RESETGROUP'] = df.groupby('ID')['_FLAG'].cumsum()
        df.drop('_FLAG', axis=1, inplace=True)

    def fn(a):
        if a[addl] == 0:
            times = [a[idv]]
            expanded = [False]
        else:
            length = int(a[addl])
            times = [a[ii] * x + a[idv] for x in range(length + 1)]
            expanded = [False] + [True] * length
        a['_TIMES'] = times
        a['_EXPANDED'] = expanded
        return a

    df = df.apply(fn, axis=1)
    df = df.apply(lambda x: x.explode() if x.name in ['_TIMES', '_EXPANDED'] else x)
    df = df.astype({'_EXPANDED': np.bool_})
    df = df.groupby([idcol, '_RESETGROUP'], group_keys=False).apply(
        lambda x: x.sort_values(by='_TIMES', kind='stable')
    )
    df[idv] = df['_TIMES']
    df.drop(['_TIMES', '_RESETGROUP'], axis=1, inplace=True)
    if flag:
        df.rename(columns={'_EXPANDED': 'EXPANDED'}, inplace=True)
    else:
        df.drop([addl, ii, '_EXPANDED'], axis=1, inplace=True)
    model = model.replace(dataset=df.reset_index(drop=True))
    return model.update_source()


def get_doseid(model: Model):
    """Get a DOSEID series from the dataset with an id of each dose period starting from 1

    If a a dose and observation exist at the same time point the observation will be counted
    towards the previous dose.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        DOSEIDs

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_doseid
    >>> model = load_example_model("pheno")
    >>> get_doseid(model)  # doctest: +ELLIPSIS
    0       1
    1       1
    2       2
    3       3
    4       4
           ..
    739    10
    740    11
    741    12
    742    13
    743    13
    Name: DOSEID, Length: 744, dtype: int...
    """
    try:
        dose = model.datainfo.typeix['dose'][0].name
    except IndexError:
        raise DatasetError('Could not identify dosing rows in dataset')

    df = model.dataset.copy()
    df['DOSEID'] = df[dose]
    df.loc[df['DOSEID'] > 0, 'DOSEID'] = 1
    df['DOSEID'] = df['DOSEID'].astype(int)
    idcol = model.datainfo.id_column.name
    df['DOSEID'] = df.groupby(idcol)['DOSEID'].cumsum()

    # Adjust for dose and observation at the same time point
    # Observation is moved to previous dose group
    # Except for steady state dose where the dose group is kept
    try:
        eventcol = model.datainfo.typeix['event'][0].name
    except IndexError:
        df['_RESETGROUP'] = 1.0
    else:
        df['_FLAG'] = df[eventcol] >= 3
        df['_RESETGROUP'] = df.groupby('ID')['_FLAG'].cumsum()

    try:
        ss = model.datainfo.typeix['ss'][0].name
    except IndexError:
        ss = None

    idvcol = model.datainfo.idv_column.name
    ser = df.groupby([idcol, idvcol, '_RESETGROUP']).size()
    nonunique = ser[ser > 1]

    for i, time, _ in nonunique.index:
        groupind = df[(df[idcol] == i) & (df[idvcol] == time)].index
        obsind = df[(df[idcol] == i) & (df[idvcol] == time) & (df[dose] == 0)].index
        doseind = set(groupind) - set(obsind)
        if not doseind:
            continue
        maxind = max(doseind)
        for index in obsind:
            if 0 in groupind:  # This is the first dose
                continue
            if maxind > index:  # Dose record is after the observation
                continue
            if ss and df.loc[maxind, ss] > 0:  # No swap for SS dosing
                continue
            curdoseid = df.loc[index, 'DOSEID']
            df.loc[index, 'DOSEID'] = curdoseid - 1

    return df['DOSEID'].copy()


def get_mdv(model: Model):
    """Get MDVs from dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        MDVs
    """
    found = False
    for key in ['mdv', 'event', 'dose']:
        try:
            label = model.datainfo.typeix[key][0].name
            found = True
            break
        except IndexError:
            pass
    else:
        label = model.datainfo.dv_column.name

    data = model.dataset[label].astype('float64').squeeze()

    series = data.where(data == 0, other=1) if found else pd.Series(np.zeros(len(data)))

    return series.astype('int32').rename('MDV')


def get_evid(model: Model):
    """Get the evid from model dataset

    If an event column is present this will be extracted otherwise
    an evid column will be created.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        EVID
    """
    di = model.datainfo
    try:
        eventcols = di.typeix['event']
    except IndexError:
        pass
    else:
        return model.dataset[eventcols[0].name]
    mdv = get_mdv(model)
    return mdv.rename('EVID')


def get_admid(model: Model):
    """Get the admid from model dataset

    If an administration column is present this will be extracted otherwise
    an admid column will be created based on the admids of the present doses.
    This is dependent on the presence of a CMT column to be generated correctly.

    When generated, admids of events in between doses is set to the last used
    admid.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        ADMID
    """
    di = model.datainfo
    try:
        admidcols = di.typeix["admid"]
    except IndexError:
        pass
    else:
        return model.dataset[admidcols[0].name]

    odes = model.statements.ode_system
    names = odes.compartment_names
    remap = {}
    if isinstance(odes, CompartmentalSystem):
        for dosing in odes.dosing_compartments:
            remap[names.index(dosing.name) + 1] = dosing.doses[0].admid
    adm = get_cmt(model)
    adm = adm.replace(remap)
    adm.name = "ADMID"

    # Replace all observations with the previous admid type
    current_admin = adm[0]
    current_subject = model.dataset["ID"][0]
    for i, data in enumerate(zip(get_evid(model), adm, model.dataset["ID"])):
        event = data[0]
        admin = data[1]
        subject = data[2]
        if current_subject == subject:
            if event == 1:
                current_admin = admin
            if event != 1:
                if current_admin is not None:
                    adm[i] = current_admin
        else:
            current_subject = subject
            current_admin = admin
    return adm


def add_admid(model: Model):
    """
    Add an admid column to the model dataset and datainfo. Dependent on the
    presence of a CMT column in order to add admid correctly.

    When generated, admids of events in between doses is set to the last used
    admid.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    model : Model
        Pharmpy model

    See also
    --------
    get_admid : Get or create an admid column
    get_cmt : Get or create a cmt column
    """
    di = model.datainfo
    if "admid" not in di.types:
        adm = get_admid(model)
        dataset = model.dataset
        dataset["ADMID"] = adm
        di = update_datainfo(model.datainfo, dataset)
        colinfo = di['ADMID'].replace(type='admid')
        model = model.replace(datainfo=di.set_column(colinfo), dataset=dataset)

    return model.update_source()


def set_admid(model: Model, column_name: str):
    """
    Set the specified column in the dataset to the admid data type.

    Parameters
    ----------
    model : Model
        Pharmpy model
    column_name : str
        name of column to set as admid

    Returns
    -------
    model : Model
        Pharmpy model

    See also
    --------
    get_admid : Get or create an admid column
    add_admid : Add an admid column to the dataset
    """
    di = model.datainfo
    colinfo = di[column_name].replace(type="admid")
    model = model.replace(datainfo=di.set_column(colinfo))
    return model.update_source()


def get_cmt(model: Model):
    """Get the cmt (compartment) column from the model dataset

    If a cmt column is present this will be extracted otherwise
    a cmt column will be created. If created, multiple dose compartments are
    dependent on the presence of an admid type column, otherwise, dose/non-dose
    will be considered.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        CMT
    """
    di = model.datainfo
    try:
        cmtcols = di.typeix['compartment']
    except IndexError:
        pass
    else:
        return model.dataset[cmtcols[0].name]

    # See if admid exist
    try:
        admidcols = di.typeix["admid"]
    except IndexError:
        # No admid found --> Assume dose/non-dose
        odes = model.statements.ode_system
        if isinstance(odes, CompartmentalSystem):
            dosing = odes.dosing_compartments[0]
            names = odes.compartment_names
            dose_cmt = names.index(dosing.name) + 1
        else:
            dose_cmt = 1
        cmt = get_evid(model)
        cmt = cmt.replace({1: dose_cmt, 2: 0, 3: 0, 4: dose_cmt})  # Only consider dose/non-dose
        cmt.name = "CMT"
        return cmt
    else:
        admidcols = model.dataset[admidcols[0].name]
        # Admid found -> convert to CMT based on doses
        odes = model.statements.ode_system
        names = odes.compartment_names
        remap = {}
        if isinstance(odes, CompartmentalSystem):
            for dosing in odes.dosing_compartments:
                if dosing == odes.central_compartment:
                    remap[2] = names.index(dosing.name) + 1
                    central_number = names.index(dosing.name) + 1
                else:
                    remap[1] = names.index(dosing.name) + 1
        admidcols = admidcols.replace(remap)
        admidcols.loc[get_evid(model) == 0] = central_number
        admidcols.name = "ADMID"
        return admidcols


def add_cmt(model: Model):
    """Add a CMT column to the model dataset and datainfo if not existed

    In case of multiple doses, this method is dependent on the presence of an
    admid column to correctly number each dose.

    NOTE : Existing CMT is based on datainfo type being set to 'compartment'
    and a column named 'CMT' can be replaced

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    model : Model
        Pharmpy model

    See also
    --------
    get_admid : Get or create an admid column
    get_cmt : Get or create a cmt column
    """
    di = model.datainfo
    if "compartment" not in di.types:
        cmt_name = "CMT"
        cmt = get_cmt(model)
        dataset = model.dataset
        dataset[cmt_name] = cmt
        di = update_datainfo(model.datainfo, dataset)
        colinfo = di[cmt_name].replace(type='compartment')
        model = model.replace(datainfo=di.set_column(colinfo), dataset=dataset)

    return model.update_source()


def add_time_after_dose(model: Model):
    """Calculate and add a TAD column to the dataset"

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, add_time_after_dose
    >>> model = load_example_model("pheno")
    >>> model = add_time_after_dose(model)

    """
    try:
        model.datainfo.descriptorix['time after dose']
    except IndexError:
        pass
    else:
        # Already have time after dose
        return model
    temp = translate_nmtran_time(model)
    idv = temp.datainfo.idv_column.name
    idlab = temp.datainfo.id_column.name
    df = model.dataset.copy()
    df['_NEWTIME'] = temp.dataset[idv]

    try:
        addl = temp.datainfo.typeix['additional'][0].name
    except IndexError:
        addl = None
    else:
        # FIXME: Temp workaround, should be canonicalized in Model.replace
        di = update_datainfo(temp.datainfo, df)
        new_idvcol = di.idv_column.replace(type='unknown')
        new_timecol = di['_NEWTIME'].replace(type='idv')
        di = di.set_column(new_idvcol).set_column(new_timecol)
        temp = temp.replace(datainfo=di, dataset=df)
        temp = expand_additional_doses(temp, flag=True)
        df = temp.dataset

    df['_DOSEID'] = get_doseid(temp)

    # Sort in case DOSEIDs are non-increasing
    df = (
        df.groupby(idlab)
        .apply(lambda x: x.sort_values(by=['_DOSEID'], kind='stable', ignore_index=True))
        .reset_index(drop=True)
    )

    df['TAD'] = df.groupby([idlab, '_DOSEID'])['_NEWTIME'].diff().fillna(0)
    df['TAD'] = df.groupby([idlab, '_DOSEID'])['TAD'].cumsum()

    if addl:
        df = df[~df['EXPANDED']].reset_index(drop=True)
        df.drop(columns=['EXPANDED'], inplace=True)

    # Handle case for observation at same timepoint as SS dose
    # In this case II should be used as TAD (imaginary previous dose)
    try:
        ss = model.datainfo.typeix['ss'][0].name
        ii = model.datainfo.typeix['ii'][0].name
    except IndexError:
        pass
    else:

        def fn(df):
            if len(df) < 2:
                return df
            ii_time = None
            for i in df.index:
                if df.loc[i, ss] > 0:
                    ii_time = df.loc[i, ii]
                else:
                    assert ii_time is not None
                    df.loc[i, 'TAD'] = ii_time
            return df

        df = df.groupby([idlab, idv, '_DOSEID'], group_keys=False).apply(fn)

    df.drop(columns=['_NEWTIME', '_DOSEID'], inplace=True)

    # FIXME: Temp workaround, should be canonicalized in Model.replace
    di = update_datainfo(model.datainfo, df)
    colinfo = di['TAD'].replace(descriptor='time after dose', unit=di[idv].unit)
    model = model.replace(datainfo=di.set_column(colinfo), dataset=df)
    return model.update_source()


def get_concentration_parameters_from_data(model: Model):
    """Create a dataframe with concentration parameters

    Note that all values are directly calculated from the dataset

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    pd.DataFrame
        Concentration parameters

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_concentration_parameters_from_data
    >>> model = load_example_model("pheno")
    >>> get_concentration_parameters_from_data(model)
               Cmax  Tmax  Cmin  Tmin
    ID DOSEID
    1  1       17.3   2.0   NaN   NaN
       2        NaN   NaN   NaN   NaN
       3        NaN   NaN   NaN   NaN
       4        NaN   NaN   NaN   NaN
       5        NaN   NaN   NaN   NaN
    ...         ...   ...   ...   ...
    59 9        NaN   NaN   NaN   NaN
       10       NaN   NaN   NaN   NaN
       11       NaN   NaN   NaN   NaN
       12       NaN   NaN   NaN   NaN
       13      40.2   2.0   NaN   NaN
    <BLANKLINE>
    [589 rows x 4 columns]
    """
    model = add_time_after_dose(model)
    doseid = get_doseid(model)
    df = model.dataset.copy()
    df['DOSEID'] = doseid
    idlab = model.datainfo.id_column.name
    dv = model.datainfo.dv_column.name
    noobs = df.groupby([idlab, 'DOSEID']).size() == 1
    idx = df.groupby([idlab, 'DOSEID'])[dv].idxmax()
    params = df.loc[idx].set_index([idlab, 'DOSEID'])
    params = params[[dv, 'TAD']]
    params.rename(columns={dv: 'Cmax', 'TAD': 'Tmax'}, inplace=True)
    params.loc[noobs] = np.nan

    grpind = df.groupby(['ID', 'DOSEID']).indices
    keep = []
    for ind, rows in grpind.items():
        index = idx.loc[ind]
        p = params.loc[ind]
        if not np.isnan(p['Tmax']):
            keep += [row for row in rows if row > index]
    minidx = df.iloc[keep].groupby([idlab, 'DOSEID'])[dv].idxmin()
    params2 = df.loc[minidx].set_index([idlab, 'DOSEID'])
    params2 = params2[[dv, 'TAD']]
    params2.rename(columns={dv: 'Cmin', 'TAD': 'Tmin'}, inplace=True)
    res = params.join(params2)
    return res


def drop_dropped_columns(model: Model):
    """Drop columns marked as dropped from the dataset

    NM-TRAN date columns will not be dropped by this function
    even if marked as dropped.
    Columns not specified in the datainfo ($INPUT for NONMEM)
    will also be dropped from the dataset.

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = drop_dropped_columns(model)
    >>> list(model.dataset.columns)
    ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV', 'FA1', 'FA2']

    See also
    --------
    drop_columns : Drop specific columns or mark them as drop
    """
    datainfo = model.datainfo
    todrop = [
        colname
        for colname in datainfo.names
        if datainfo[colname].drop and datainfo[colname].datatype != 'nmtran-date'
    ]
    todrop += list(set(model.dataset.columns) - set(datainfo.names))
    model = drop_columns(model, todrop)
    return model.update_source()


def drop_columns(model: Model, column_names: Union[List[str], str], mark: bool = False):
    """Drop columns from the dataset or mark as dropped

    Parameters
    ----------
    model : Model
        Pharmpy model object
    column_names : list or str
        List of column names or one column name to drop or mark as dropped
    mark : bool
        Default is to remove column from dataset. Set this to True to only mark as dropped

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = drop_columns(model, ['WGT', 'APGR'])
    >>> list(model.dataset.columns)
    ['ID', 'TIME', 'AMT', 'DV', 'FA1', 'FA2']

    See also
    --------
    drop_dropped_columns : Drop all columns marked as drop
    undrop_columns : Undrop columns of model
    """
    if isinstance(column_names, str):
        column_names = [column_names]
    di = model.datainfo
    newcols, to_drop = [], []
    for col in di:
        if col.name in column_names:
            if mark:
                newcol = col.replace(drop=True)
                newcols.append(newcol)
            else:
                to_drop.append(col.name)
        else:
            newcols.append(col)
    replace_dict = {'datainfo': di.replace(columns=newcols)}
    if to_drop:
        df = model.dataset.copy()
        replace_dict['dataset'] = df.drop(to_drop, axis=1)
    model = model.replace(**replace_dict)
    return model.update_source()


def undrop_columns(model: Model, column_names: Union[List[str], str]):
    """Undrop columns of model

    Parameters
    ----------
    model : Model
        Pharmpy model object
    column_names : list or str
        List of column names or one column name to undrop

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = drop_columns(model, ['WGT', 'APGR'], mark=True)
    >>> model = undrop_columns(model, 'WGT')

    See also
    --------
    drop_dropped_columns : Drop all columns marked as drop
    drop_columns : Drop or mark columns as dropped
    """
    if isinstance(column_names, str):
        column_names = [column_names]
    di = model.datainfo
    newcols = []
    for col in di:
        if col.name in column_names:
            newcol = col.replace(drop=False)
            newcols.append(newcol)
        else:
            newcols.append(col)
    model = model.replace(datainfo=di.replace(columns=newcols))
    return model.update_source()


def _translate_nonmem_time_value(time):
    if ':' in time:
        components = time.split(':')
        if len(components) != 2:
            raise DatasetError(f'Bad TIME format: {time}')
        hours = float(components[0]) + float(components[1]) / 60
        return hours
    else:
        return float(time)


def _translate_time_column(df, timecol, idcol):
    if df[timecol].dtype != np.float64:
        df[timecol] = df[timecol].apply(_translate_nonmem_time_value)
        df[timecol] = df[timecol] - df.groupby(idcol)[timecol].transform('first')
    return df


def _translate_nonmem_time_and_date_value(ser, timecol, datecol):
    timeval = _translate_nonmem_time_value(ser[timecol])
    date = ser[datecol]
    a = re.split(r'[^0-9]', date)
    if date.startswith('-') or len(a) == 1:
        return timeval + float(date) * 24
    elif len(a) == 2:
        year = 2001  # Non leap year
        month = a[1]
        day = a[0]
    elif len(a) == 3:
        if datecol.endswith('E'):
            month = a[0]
            day = a[1]
            year = a[2]
        elif datecol.endswith('1'):
            day = a[0]
            month = a[1]
            year = a[2]
        elif datecol.endswith('3'):
            year = a[0]
            day = a[1]
            month = a[2]
        else:  # Let DAT2 be default if other name
            year = a[0]
            month = a[1]
            day = a[2]
        if len(year) < 3:
            year = int(year)
            if year > 50:
                year += 1900
            else:
                year += 2000
        else:
            year = int(year)
        month = int(month)
        day = int(day)
        hour = int(timeval)
        timeval = (timeval - hour) * 60
        minute = int(timeval)
        timeval = (timeval - minute) * 60
        second = int(timeval)
        timeval = (timeval - second) * 1000000
        microsecond = int(timeval)
        timeval = (timeval - microsecond) * 1000
        nanosecond = int(timeval)
        ts = pd.Timestamp(
            year=year,
            month=month,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond,
            nanosecond=nanosecond,
        )
        return ts
    else:
        raise DatasetError(f'Bad DATE value: {date}')


def _translate_time_and_date_columns(df, timecol, datecol, idcol):
    df[timecol] = df.apply(
        _translate_nonmem_time_and_date_value, axis=1, timecol=timecol, datecol=datecol
    )
    timediff = df[timecol] - df.groupby(idcol)[timecol].transform('first')
    if df[timecol].dtype != np.float64:
        df[timecol] = timediff.dt.total_seconds() / 3600
    return df


def _find_time_and_date_columns(model):
    # Both time and date can be None. If date is None time must be not None
    time = None
    date = None
    di = model.datainfo
    for col in di:
        if col.datatype == 'nmtran-time' and not col.drop:
            if time is None:
                time = col
            else:
                raise ValueError(f"Multiple time columns found {time} and {col.name}")
        elif col.datatype == 'nmtran-date' and not col.drop:
            if date is None:
                date = col
            else:
                raise ValueError(f"Multiple date columns found {date} and {col.name}")
    if time is None and date is not None:
        raise ValueError(f"Found date column {date}, but no time column")
    return time, date


def translate_nmtran_time(model: Model):
    """Translate NM-TRAN TIME and DATE column into one TIME column

    If dataset of model have special NM-TRAN TIME and DATE columns these
    will be translated into one single time column with time in hours.

    Warnings
    --------
    Use this function with caution. For example reset events are currently not taken into account.

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Pharmpy model object
    """
    timecol, datecol = _find_time_and_date_columns(model)
    df = model.dataset.copy()
    di = model.datainfo
    idname = di.id_column.name
    if datecol is None:
        if timecol is None:
            return model
        else:
            df = _translate_time_column(df, timecol.name, idname)
    else:
        assert timecol is not None
        df = _translate_time_and_date_columns(df, timecol.name, datecol.name, idname)
        model = drop_columns(model, datecol.name)
        timecol = timecol.replace(unit='h')
    timecol = timecol.replace(datatype='float64')
    di = di.set_column(timecol)
    model = model.replace(datainfo=di, dataset=df)
    return model.update_source()


def _loq_mask(
    model: Model,
    lloq: Optional[Union[float, str]] = None,
    uloq: Optional[Union[float, str]] = None,
    blq: Optional[str] = None,
    alq: Optional[str] = None,
):
    """Boolean series with False for lloq records and True for non-lloq
    Options as remove_loq_data
    """
    if blq and lloq:
        raise ValueError("Cannot specify blq and lloq at the same time")
    if alq and uloq:
        raise ValueError("Cannot specify alq and uloq at the same time")
    df = model.dataset
    if lloq is not None or uloq is not None:
        dv = model.datainfo.dv_column.name
    mdv = get_mdv(model)
    which_keep = pd.Series(True, index=df.index)
    if isinstance(lloq, str):
        lloq = df[lloq]
    if isinstance(uloq, str):
        uloq = df[uloq]
    if lloq is not None:
        which_keep &= (df[dv] > lloq) | mdv
    elif blq is not None:
        which_keep &= (df[blq] == 0) | mdv
    if uloq is not None:
        which_keep &= (df[dv] < uloq) | mdv
    elif alq is not None:
        which_keep &= (df[alq] == 0) | mdv
    return which_keep


def remove_loq_data(
    model: Model,
    lloq: Optional[Union[float, str]] = None,
    uloq: Optional[Union[float, str]] = None,
    blq: Optional[str] = None,
    alq: Optional[str] = None,
    keep: Optional[int] = 0,
):
    """Remove loq data records from the dataset

    Does nothing if none of the limits are specified.

    Parameters
    ----------
    model : Model
        Pharmpy model object
    lloq : float or str
        Value or column name for lower limit of quantification.
    uloq : float or str
        Value or column name for upper limit of quantification.
    blq : str
        Column name for below limit of quantification indicator.
    alq : str
        Column name for above limit of quantification indicator.
    keep : int
        Number of loq records to keep for each run of consecutive loq records.

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = remove_loq_data(model, lloq=10, uloq=40)
    >>> len(model.dataset)
    736

    See also
    --------
    set_lloq_data
    transform_blq

    """
    which_keep = _loq_mask(model, lloq=lloq, uloq=uloq, blq=blq, alq=alq)
    df = model.dataset
    if keep > 0:
        idcol = model.datainfo.id_column.name
        keep_df = pd.DataFrame(
            {'ID': df[idcol], 'consec': (~which_keep).diff().ne(0).cumsum(), 'remove': ~which_keep}
        )
        obj = keep_df.groupby([idcol, 'consec']).cumsum().le(keep)['remove']
        which_keep = obj | which_keep
    model = model.replace(dataset=df[which_keep])
    return model.update_source()


def set_lloq_data(
    model: Model,
    value: Union[str, float, sympy.Expr],
    lloq: Optional[Union[float, str]] = None,
    blq: Optional[str] = None,
):
    """Set a dv value for lloq data records

    Parameters
    ----------
    model : Model
        Pharmpy model object
    value : float or sympy.Expression
        The new dv value
    lloq : float or str
        Value or column name for lower limit of quantification.
    blq : str
        Column name for below limit of quantification indicator.

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_lloq_data(model, 0, lloq=10)

    See also
    --------
    remove_loq_data
    transform_blq

    """
    which_keep = _loq_mask(model, lloq=lloq, blq=blq)
    df = model.dataset.copy()
    dv = model.datainfo.dv_column.name
    if isinstance(value, sympy.Expr) or isinstance(value, str):
        value = df.eval(str(value))
    df[dv] = df[dv].where(which_keep, value)
    model = model.replace(dataset=df)
    return model


def set_reference_values(model: Model, refs: Dict[str, Union[int, float]]):
    """Set reference values for selected columns

        All values for each selected column will be replaced. For dose columns
        only the values for dosing events will be replaced.

    Parameters
    ----------
    model : Model
        Pharmpy model object
    refs : dict
        Pairs of column names and reference values

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_reference_values(model, {'WGT': 0.5, 'AMT': 4.0})
    >>> model.dataset
         ID   TIME  AMT  WGT  APGR    DV  FA1  FA2
    0     1    0.0  4.0  0.5   7.0   0.0  1.0  1.0
    1     1    2.0  0.0  0.5   7.0  17.3  0.0  0.0
    2     1   12.5  4.0  0.5   7.0   0.0  1.0  1.0
    3     1   24.5  4.0  0.5   7.0   0.0  1.0  1.0
    4     1   37.0  4.0  0.5   7.0   0.0  1.0  1.0
    ..   ..    ...  ...  ...   ...   ...  ...  ...
    739  59  108.3  4.0  0.5   6.0   0.0  1.0  1.0
    740  59  120.5  4.0  0.5   6.0   0.0  1.0  1.0
    741  59  132.3  4.0  0.5   6.0   0.0  1.0  1.0
    742  59  144.8  4.0  0.5   6.0   0.0  1.0  1.0
    743  59  146.8  0.0  0.5   6.0  40.2  0.0  0.0
    <BLANKLINE>
    [744 rows x 8 columns]

    """
    df = model.dataset
    di = model.datainfo
    newcols = dict()
    dtypes = dict()
    for colname, value in refs.items():
        if di[colname].type == 'dose':
            newdose = df[colname].mask(df[colname] > 0, value)
            newcols[colname] = newdose
        else:
            newcols[colname] = value
        datatype = ColumnInfo.convert_datatype_to_pd_dtype(di[colname].datatype)
        dtypes[colname] = datatype
    df = df.assign(**newcols).astype(dtypes)
    model = model.replace(dataset=df)
    return model


class Checker:
    _all_checks = (
        ('A1', 'Body weight has unit'),
        ('A2', 'Body weight has mass unit'),
        ('A3', 'Body weight >0 and <700kg'),
        ('A4', 'Age has unit'),
        ('A5', 'Age has time unit'),
        ('A6', 'Age >=0 and <130 years'),
        ('A7', 'Lean body mass has unit'),
        ('A8', 'Lean body mass has mass unit'),
        ('A9', 'Lean body mass >0 and <700kg'),
        ('A10', 'Fat free mass has unit'),
        ('A11', 'Fat free mass has mass unit'),
        ('A12', 'Fat free mass >0 and <700kg'),
        ('D1', 'Time after dose has unit'),
        ('D2', 'Time after dose has time unit'),
        ('D3', 'Time after dose >=0'),
        ('D4', 'Plasma concentration has unit'),
        ('D5', 'Plasma concentration has mass/volume unit'),
        ('D6', 'Plasma concentration >= 0'),
        ('I1', 'Subject identifier is unitless'),
    )

    def __init__(self, datainfo, dataset, verbose=False):
        self.datainfo = datainfo
        self.dataset = dataset
        self.verbose = verbose
        self.check_results = {}
        self.violations = []

    def set_result(self, code, test=False, violation=None, skip=False, warn=False):
        if skip:
            result = "SKIP"
        elif test:
            result = "OK"
        else:
            if warn:
                result = "WARN"
            else:
                result = "FAIL"
        if code not in self.check_results or (
            code in self.check_results
            and (
                self.check_results[code] == 'SKIP'
                or self.check_results[code] == 'OK'
                and result in ('WARN', 'FAIL')
                or self.check_results[code] == 'WARN'
                and result == 'FAIL'
            )
        ):
            self.check_results[code] = result
        if result in ('WARN', 'FAIL'):
            self.violations.append((code, result, violation))

    def check_has_unit(self, code, col):
        has_unit = col.unit is not None
        self.set_result(code, test=has_unit, violation=col.name, warn=True)
        return has_unit

    def check_is_unitless(self, code, col):
        is_unitless = col.unit == sympy.Integer(1)
        self.set_result(code, test=is_unitless, violation=col.name, warn=True)

    def check_dimension(self, code, column, dim):
        if column.unit is None:
            self.set_result(code, skip=True)
            return False
        else:
            dim2 = sympy.physics.units.Dimension(
                sympy.physics.units.si.SI.get_dimensional_expr(column.unit)
            )
            self.set_result(
                code,
                test=dim == dim2,
                violation=f"Unit {column.unit} of {column.name} is not a {dim} unit",
            )
            return dim == dim2

    def check_range(self, code, col, lower, upper, unit, lower_included=True, upper_included=True):
        name = col.name
        if lower == 0:
            scaled_lower = lower
        else:
            scaled_lower = float(sympy.physics.units.convert_to(lower * unit, col.unit) / col.unit)
        if upper == 0:
            scaled_upper = upper
        else:
            scaled_upper = float(sympy.physics.units.convert_to(upper * unit, col.unit) / col.unit)
        if lower_included:
            lower_viol = self.dataset[name] < scaled_lower
        else:
            lower_viol = self.dataset[name] <= scaled_lower
        if upper_included:
            upper_viol = self.dataset[name] > scaled_upper
        else:
            upper_viol = self.dataset[name] >= scaled_upper
        all_viol = lower_viol | upper_viol
        violations = all_viol[all_viol]
        if not violations.empty:
            for i in violations.index:
                self.set_result(
                    code,
                    test=False,
                    violation=f"{col.name} index={i} value={self.dataset[name].loc[i]}",
                )
        else:
            self.set_result(code, test=True)

    def get_dataframe(self):
        codes = []
        checks = []
        results = []
        violations = []

        for code, msg in Checker._all_checks:
            if code not in self.check_results:
                self.check_results[code] = "SKIP"

            if self.check_results[code] in ['OK', 'SKIP']:
                if self.verbose:
                    codes.append(code)
                    checks.append(msg)
                    results.append(self.check_results[code])
                    violations.append(None)
            else:
                for viol in self.violations:
                    if (
                        viol[0] == code
                        and viol[1] == "FAIL"
                        or (viol[1] == "WARN" and self.verbose)
                    ):
                        codes.append(code)
                        checks.append(msg)
                        results.append(viol[1])
                        violations.append(viol[2])
        df = pd.DataFrame(
            {'code': codes, 'check': checks, 'result': results, 'violation': violations}
        )
        return df

    def print(self):
        table = rich_table.Table(title="Dataset checks", box=rich_box.SQUARE)
        table.add_column("Code")
        table.add_column("Check")
        table.add_column("Result")
        table.add_column("Violation")

        for code, msg in Checker._all_checks:
            if code not in self.check_results:
                self.check_results[code] = "SKIP"

            if self.check_results[code] in ['OK', 'SKIP']:
                if self.verbose:
                    table.add_row(code, msg, f'[bold green]{self.check_results[code]}', "")
            else:
                for viol in self.violations:
                    if (
                        viol[0] == code
                        and viol[1] == "FAIL"
                        or (viol[1] == "WARN" and self.verbose)
                    ):
                        result = viol[1]
                        if result == "FAIL":
                            result = f"[bold red]{result}"
                        else:
                            result = f"[bold yellow]{result}"
                        table.add_row(code, msg, result, viol[2])

        if table.rows:  # Do not print an empty table
            console = rich_console.Console()
            console.print(table)


def check_dataset(model: Model, dataframe: bool = False, verbose: bool = False):
    """Check dataset for consistency across a set of rules

    Parameters
    ----------
    model : Model
        Pharmpy model object
    dataframe : bool
        True to return a DataFrame instead of printing to the console
    verbose : bool
        Print out all rules checked if True else print only failed rules

    Returns
    -------
    pd.DataFrame
        Only returns a DataFrame is dataframe=True
    """
    di = model.datainfo
    df = model.dataset
    checker = Checker(di, df, verbose=verbose)

    for col in di:
        if col.descriptor == "body weight":
            checker.check_has_unit("A1", col)
            samedim = checker.check_dimension("A2", col, sympy.physics.units.mass)
            if samedim:
                checker.check_range("A3", col, 0, 700, sympy.physics.units.kg, False, False)

        if col.descriptor == "age":
            checker.check_has_unit("A4", col)
            samedim = checker.check_dimension("A5", col, sympy.physics.units.time)
            if samedim:
                checker.check_range("A6", col, 0, 130, sympy.physics.units.year, True, False)

        if col.descriptor == "lean body mass":
            checker.check_has_unit("A7", col)
            samedim = checker.check_dimension("A8", col, sympy.physics.units.mass)
            if samedim:
                checker.check_range("A9", col, 0, 700, sympy.physics.units.kg, False, False)

        if col.descriptor == "fat free mass":
            checker.check_has_unit("A10", col)
            samedim = checker.check_dimension("A11", col, sympy.physics.units.mass)
            if samedim:
                checker.check_range("A12", col, 0, 700, sympy.physics.units.kg, False, False)

        if col.descriptor == "time after dose":
            checker.check_has_unit("D1", col)
            samedim = checker.check_dimension("D2", col, sympy.physics.units.time)
            if samedim:
                checker.check_range(
                    "D3", col, 0, float('inf'), sympy.physics.units.second, True, False
                )

        if col.descriptor == "plasma concentration":
            checker.check_has_unit("D4", col)
            samedim = checker.check_dimension(
                "D5", col, sympy.physics.units.mass / sympy.physics.units.length**3
            )
            if samedim:
                checker.check_range(
                    "D6",
                    col,
                    0,
                    float('inf'),
                    sympy.physics.units.kg / sympy.physics.units.L,
                    True,
                    False,
                )

        if col.descriptor == "subject identifier":
            checker.check_is_unitless("I1", col)

    if dataframe:
        return checker.get_dataframe()
    else:
        checker.print()


def read_dataset_from_datainfo(
    datainfo: Union[DataInfo, Path, str], datatype: Optional[str] = None
):
    """Read a dataset given a datainfo object or path to a datainfo file

    Parameters
    ----------
    datainfo : DataInfo | Path | str
        A datainfo object or a path to a datainfo object
    datatype : str
        A string to specify dataset type

    Returns
    -------
    pd.DataFrame
        The dataset
    """
    if not isinstance(datainfo, DataInfo):
        datainfo = DataInfo.read_json(datainfo)

    if datainfo.path is None:
        raise ValueError('datainfo.path is None')
    from pharmpy.model.external.nonmem.dataset import read_nonmem_dataset
    from pharmpy.model.external.nonmem.parsing import filter_observations

    if datatype == 'nonmem':
        drop = [col.drop for col in datainfo]
        df = read_nonmem_dataset(
            datainfo.path,
            ignore_character='@',
            drop=drop,
            colnames=datainfo.names,
            dtype=datainfo.get_dtype_dict(),
        )
        # This assumes a PK model
        df = filter_observations(df, list(df.columns))
    else:
        df = pd.read_csv(
            datainfo.path,
            sep=datainfo.separator,
            dtype=datainfo.get_dtype_dict(),
            float_precision='round_trip',
        )
    return df


def deidentify_data(
    df: pd.DataFrame, id_column: str = 'ID', date_columns: Optional[List[str]] = None
):
    """Deidentify a dataset

    Two operations are performed on the dataset:

    1. All ID numbers are randomized from the range 1 to n
    2. All columns containing dates will have the year changed

    The year change is done by letting the earliest year in the dataset
    be used as a reference and by maintaining leap years. The reference year
    will either be 1901, 1902, 1903 or 1904 depending on its distance to the closest
    preceeding leap year.

    Parameters
    ----------
    df : pd.DataFrame
        A dataset
    id_column : str
        Name of the id column
    date_columns : list
        Names of all date columns

    Returns
    -------
    pd.DataFrame
        Deidentified dataset
    """
    df = df.copy()
    df[id_column] = pd.to_numeric(df[id_column])
    resampler = resample_data(df, id_column)
    df, _ = next(resampler)

    if date_columns is None:
        return df
    for datecol in date_columns:
        if pd.api.types.is_datetime64_any_dtype(df[datecol]):
            pass
        elif df[datecol].dtype == 'object':
            # assume string
            df[datecol] = pd.to_datetime(df[datecol])
        else:
            raise ValueError(f"Column {datecol} does not seem to contain a date")
    earliest_date = df[date_columns].min().min()

    # Handle leap year modulo
    earliest_year_modulo = earliest_date.year % 4
    reference_offset = 4 if earliest_year_modulo == 0 else earliest_year_modulo
    reference_year = 1900 + reference_offset
    delta = earliest_date.year - reference_year

    def convert(x):
        new = x.replace(year=x.year - delta)
        return new

    for datecol in date_columns:
        df[datecol] = df[datecol].transform(convert)

    return df


def unload_dataset(model: Model):
    """Unload the dataset from a model

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        Pharmpy model with dataset removed

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, unload_dataset
    >>> model = load_example_model("pheno")
    >>> model = unload_dataset(model)
    >>> model.dataset is None
    True

    """
    model = model.replace(dataset=None)
    return model


def load_dataset(model: Model):
    """Load the dataset given datainfo

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        Pharmpy model with dataset removed

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, load_dataset, unload_dataset
    >>> model = load_example_model("pheno")
    >>> model = unload_dataset(model)
    >>> model.dataset is None
    True
    >>> model = load_dataset(model)
    >>> model.dataset
         ID   TIME   AMT  WGT  APGR    DV  FA1  FA2
    0     1    0.0  25.0  1.4   7.0   0.0  1.0  1.0
    1     1    2.0   0.0  1.4   7.0  17.3  0.0  0.0
    2     1   12.5   3.5  1.4   7.0   0.0  1.0  1.0
    3     1   24.5   3.5  1.4   7.0   0.0  1.0  1.0
    4     1   37.0   3.5  1.4   7.0   0.0  1.0  1.0
    ..   ..    ...   ...  ...   ...   ...  ...  ...
    739  59  108.3   3.0  1.1   6.0   0.0  1.0  1.0
    740  59  120.5   3.0  1.1   6.0   0.0  1.0  1.0
    741  59  132.3   3.0  1.1   6.0   0.0  1.0  1.0
    742  59  144.8   3.0  1.1   6.0   0.0  1.0  1.0
    743  59  146.8   0.0  1.1   6.0  40.2  0.0  0.0
    <BLANKLINE>
    [744 rows x 8 columns]

    """
    df = read_dataset_from_datainfo(model.datainfo)
    model = model.replace(dataset=df)
    return model


def bin_observations(
    model: Model, method: Literal["equal_width", "equal_number"], nbins: int
) -> pd.Series:
    """Bin all observations on the independent variable

    Available binning methods:

    +---------------+-------------------------------------------------+
    | Method        | Description                                     |
    +===============+=================================================+
    | equal_width   | Bins with equal width based on the idv          |
    +---------------+-------------------------------------------------+
    | equal_number  | Bins containing an equal number of observations |
    +---------------+-------------------------------------------------+

    Parameters
    ----------
    model : Model
        Pharmpy model
    method : str
        Name of the binning method to use
    nbins : int
        The number of binns wanted

    Returns
    -------
    pd.Series
        A series of bin ids indexed on the original record index of the dataset

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, bin_observations
    >>> model = load_example_model("pheno")
    >>> bin_observations(model, method="equal_width", nbins=10)
    421     1
    527     1
    118     1
    135     1
    512     1
           ..
    203     8
    475     8
    510     8
    133     9
    267    10
    Length: 155, dtype: int32
    """

    observations = get_observations(model, keep_index=True)
    obs = model.dataset.loc[observations.index]
    idv = model.datainfo.idv_column.name
    sorted_idvs = obs[idv].sort_values()
    method_lower = method.lower()
    if method_lower == "equal_width":
        min_idv = sorted_idvs.iloc[0]
        max_idv = sorted_idvs.iloc[-1]
        step = (max_idv - min_idv) / nbins
        next_cutoff = min_idv + step
        binno = 1
        bincol = pd.Series(0, index=sorted_idvs.index, dtype='int32')
        for i in range(len(sorted_idvs)):
            if sorted_idvs.iloc[i] > next_cutoff:
                next_cutoff += step
                binno += 1
            bincol.iloc[i] = binno
    elif method_lower == "equal_number":
        bincol = pd.Series(0, index=sorted_idvs.index, dtype='int32')
        nobs = len(sorted_idvs)
        i = 0
        binno = 1
        while nobs > 0:
            n = round(nobs / nbins)
            bincol.iloc[i : i + n] = binno
            binno += 1
            i += n
            nobs -= n
            nbins -= 1
    else:
        raise ValueError(f"Unknown binning method {method}")
    return bincol



from .ml import predict_influential_individuals, predict_influential_outliers, predict_outliers
from .summarize_individuals import summarize_individuals, summarize_individuals_count_table

__all__ = [
    'predict_influential_individuals',
    'predict_influential_outliers',
    'predict_outliers',
    'summarize_individuals',
    'summarize_individuals_count_table',
]



import pytest

from pharmpy.model import Assignment
from pharmpy.modeling import add_allometry, add_peripheral_compartment, set_tmdd


def test_allometry(load_model_for_test, testdata):
    ref_model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')

    model = add_allometry(
        ref_model,
        allometric_variable='WGT',
        reference_value=70,
        parameters=['CL'],
        initials=[0.7],
        lower_bounds=[0],
        upper_bounds=[2],
        fixed=True,
    )
    assert model.statements[1] == Assignment.create('CL', 'CL*(WGT/70)**ALLO_CL')
    assert model.parameters['ALLO_CL'].init == 0.7
    assert model.parameters['ALLO_CL'].lower == 0
    assert model.parameters['ALLO_CL'].upper == 2
    with pytest.raises(ValueError):
        add_allometry(
            model,
            allometric_variable='WGT',
            reference_value=70,
            parameters=['CL'],
            initials=[0.7, 23],
            lower_bounds=[0],
            upper_bounds=[2],
            fixed=True,
        )

    with pytest.raises(ValueError):
        add_allometry(
            model,
            allometric_variable='WGT',
            reference_value=70,
            parameters=['CL'],
            initials=[0.7],
            lower_bounds=[1, 2],
            upper_bounds=[2],
            fixed=True,
        )

    with pytest.raises(ValueError):
        add_allometry(
            model,
            allometric_variable='WGT',
            reference_value=70,
            parameters=['CL'],
            initials=[0.7],
            lower_bounds=[1],
            upper_bounds=[1, 2],
            fixed=True,
        )

    with pytest.raises(ValueError):
        add_allometry(model, allometric_variable='WGT', reference_value=70, parameters=[])

    model = add_allometry(
        ref_model,
        allometric_variable='WGT',
        reference_value=70,
        parameters=['CL', 'V'],
        initials=[0.7, 0.8],
        lower_bounds=[0, 0.1],
        upper_bounds=[2, 3],
        fixed=True,
    )
    assert model.statements[1] == Assignment.create('CL', 'CL*(WGT/70)**ALLO_CL')
    assert model.statements[3] == Assignment.create('V', 'V*(WGT/70)**ALLO_V')
    assert model.parameters['ALLO_V'].init == 0.8

    model = add_allometry(
        ref_model, allometric_variable='WGT', reference_value=70, parameters=['CL', 'V']
    )
    assert model.parameters['ALLO_CL'].init == 0.75
    assert model.parameters['ALLO_V'].init == 1

    model = add_allometry(ref_model, allometric_variable='WGT', reference_value=70)
    assert model.parameters['ALLO_CL'].init == 0.75
    assert model.parameters['ALLO_V'].init == 1

    model = add_peripheral_compartment(ref_model)
    model = add_allometry(model, allometric_variable='WGT', reference_value=70)
    assert model.statements[1] == Assignment.create('VP1', 'VP1*(WGT/70)**ALLO_VP1')
    assert model.statements[3] == Assignment.create('QP1', 'QP1*(WGT/70)**ALLO_QP1')
    assert model.statements[5] == Assignment.create('CL', 'CL*(WGT/70)**ALLO_CL')
    assert model.statements[7] == Assignment.create('V1', 'V1*(WGT/70)**ALLO_V1')
    assert model.parameters['ALLO_VP1'].init == 1.0
    assert model.parameters['ALLO_QP1'].init == 0.75
    model = add_peripheral_compartment(model)
    model = add_allometry(model, allometric_variable='WGT', reference_value=70)
    assert model.statements[1] == Assignment.create('VP2', 'VP2*(WGT/70)**ALLO_VP2')
    assert model.statements[3] == Assignment.create('QP2', 'QP2*(WGT/70)**ALLO_QP2')
    assert model.parameters['ALLO_VP2'].init == 1.0
    assert model.parameters['ALLO_QP2'].init == 0.75

    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_trans1.mod')
    with pytest.raises(ValueError):
        add_allometry(model, allometric_variable='WGT', reference_value=70)

    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan3.mod')
    model = add_allometry(model, allometric_variable='WGT', reference_value=70)
    assert model.parameters['ALLO_Q'].init == 0.75
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_advan3_trans1.mod')
    with pytest.raises(ValueError):
        add_allometry(model, allometric_variable='WGT', reference_value=70)

    add_allometry(
        model,
        allometric_variable='WGT',
        reference_value=70,
        parameters=['K'],
        initials=[1],
        lower_bounds=[0],
        upper_bounds=[1],
    )


def test_add_allometry_tmdd(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="full")
    add_allometry(
        model,
        allometric_variable='WGT',
        reference_value=70,
    )

    # FIXME: Currently adds to CL and LAFREE, is this correct?
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="qss")
    add_allometry(
        model,
        allometric_variable='WGT',
        reference_value=70,
    )



import pytest


@pytest.mark.parametrize(
    "buf,advan,trans",
    [
        ('$SUBROUTINES ADVAN1 TRANS2', 'ADVAN1', 'TRANS2'),
        ('$SUBROUTINES ADVAN15', 'ADVAN15', 'TRANS1'),
        ('$SUBROUTINES ADVAN=ADVAN2', 'ADVAN2', 'TRANS1'),
        ('$SUBROUTINES ADVAN=ADVAN5 TRANS=TRANS1', 'ADVAN5', 'TRANS1'),
    ],
)
def test_advan_trans(parser, buf, advan, trans):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert rec.advan == advan
    assert rec.trans == trans



import textwrap

import pytest

from pharmpy.internals.parse import AttrToken, AttrTree, prettyprint


def assert_create(expect, *args, **kwargs):
    """Help asserter for AttrTree.create tests."""

    obj = AttrTree.create(*args, **kwargs)
    assert isinstance(obj, AttrTree)
    for item in obj.children:
        if 'tree' in item.rule:
            assert isinstance(item, AttrTree)
        elif 'LEAF' in item.rule:
            assert isinstance(item, AttrToken)
    pretty = str(prettyprint.transform(obj, content=True))
    expect = textwrap.dedent(expect).strip()
    assert pretty == expect


def test_tree_create_bad():
    """Test expected raising for planned non-legality."""

    # tree can't be empty (that's a leaf, i.e. a token)
    with pytest.raises(TypeError):  # "not iterable"
        AttrTree.create('root', None)
    with pytest.raises(ValueError):
        AttrTree.create('root', [])
    with pytest.raises(ValueError):
        AttrTree.create('root', {})

    # tokens must have str as only "child", but trees can't have str as "children"
    with pytest.raises(TypeError):  # "not iterable"
        AttrTree.create('root', 'str is only for tokens')


def test_tree_create_shallow():
    """Test creating shallow trees."""

    out = """
    tree "content"
     └─ __ANON_1 "content"
    """
    assert_create(out, 'tree', ['content'])
    out = """
    __ANON_0 ""
     └─ __ANON_1 ""
    """
    assert_create(out, '', [''])
    out = """
    __ANON_0 "LEAF"
     ├─ __ANON_1 ""
     ├─ __ANON_2 "LEAF"
     └─ __ANON_3 ""
    """
    assert_create(out, '', ['', 'LEAF', ''])

    out = """
    root "LEAF"
     └─ A_TOKEN "LEAF"
    """
    assert_create(out, 'root', dict(A_TOKEN='LEAF'))
    out = """
    __ANON_0 ""
     └─ LEAF ""
    """
    assert_create(out, '', dict(LEAF=''))


def test_tree_create_deep():
    """Test creating trees with deep, actually usable, structures."""

    inp = {
        'firstLEAF': '(leaf #1) ',
        'tree_anons': ['TEXT123 ', "'some string maybe' "],
        'tree': dict(nested_tree=dict(end_LEAF_node='!?#@')),
        'top_tree_again': dict(INTLEAF=123.456),
        'LEAF': 'THE_END',
    }
    out = """
    root "(leaf #1) TEXT123 'some string maybe' !?#@123.456THE_END"
     ├─ firstLEAF "(leaf #1) "
     ├─ tree_anons "TEXT123 'some string maybe' "
     │  ├─ __ANON_1 "TEXT123 "
     │  └─ __ANON_2 "'some string maybe' "
     ├─ tree "!?#@"
     │  └─ nested_tree "!?#@"
     │     └─ end_LEAF_node "!?#@"
     ├─ top_tree_again "123.456"
     │  └─ INTLEAF "123.456"
     └─ LEAF "THE_END"
    """
    assert_create(out, 'root', inp)


def test_tree_create_abuse():
    """Test weird abuse of the creator."""

    # It flattens to anonymous leaves? That.. is neat!
    out = """
    __ANON_0 "LEAF"
     └─ __ANON_1 "LEAF"
    """
    assert_create(out, None, [['LEAF']])
    out = """
    __ANON_0 "1234"
     ├─ __ANON_1 "1"
     ├─ __ANON_2 "2"
     ├─ __ANON_3 "3"
     └─ __ANON_4 "4"
    """
    assert_create(out, False, [[['1'], ['2']], ['3'], '4'])

    # just throwing stuff at the wall.. but it seems to stick!
    od = {'good_tree': dict(LEAF_A=' (^._.^)~ hello! '), 'bad_tree': dict(LEAF_B=None)}
    inp = {
        'item': [od, dict(dict(Btree=[dict(END_LEAF='...THE END')])), dict(_LEAF_=' (nope, here!)')]
    }
    out = """
    root " (^._.^)~ hello! None...THE END (nope, here!)"
     └─ item " (^._.^)~ hello! None...THE END (nope, here!)"
        ├─ good_tree " (^._.^)~ hello! "
        │  └─ LEAF_A " (^._.^)~ hello! "
        ├─ bad_tree "None"
        │  └─ LEAF_B "None"
        ├─ Btree "...THE END"
        │  └─ END_LEAF "...THE END"
        └─ _LEAF_ " (nope, here!)"
    """
    assert_create(out, 'root', inp)



from functools import partial
from typing import Iterable

from pharmpy.modeling import add_effect_compartment

from ..statement.feature.effect_comp import EffectComp
from ..statement.feature.symbols import Name, Wildcard
from ..statement.statement import Statement
from .feature import Feature


def features(model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, EffectComp):
            modes = (
                [Name('LINEAR'), Name('EMAX'), Name('SIGMOID')]
                if isinstance(statement.modes, Wildcard)
                else statement.modes
            )
            for mode in modes:
                yield ('EFFECTCOMP', mode.name), partial(
                    add_effect_compartment, expr=mode.name.lower()
                )



from collections import Counter, defaultdict
from dataclasses import astuple, dataclass, replace
from itertools import count
from typing import Any, Callable, Iterable, List, Literal, Optional, Sequence, Tuple, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import get_pk_parameters, has_covariate_effect, remove_covariate_effect
from pharmpy.modeling.covariate_effect import get_covariates_allowed_in_covariate_effect
from pharmpy.modeling.lrt import best_of_many as lrt_best_of_many
from pharmpy.modeling.lrt import p_value as lrt_p_value
from pharmpy.modeling.lrt import test as lrt_test
from pharmpy.tools import is_strictness_fulfilled, summarize_modelfit_results
from pharmpy.tools.common import create_results, update_initial_estimates
from pharmpy.tools.mfl.feature.covariate import EffectLiteral, InputSpec, all_covariate_effects
from pharmpy.tools.mfl.feature.covariate import features as covariate_features
from pharmpy.tools.mfl.feature.covariate import parse_spec, spec
from pharmpy.tools.mfl.parse import parse as mfl_parse
from pharmpy.tools.mfl.statement.feature.covariate import Covariate
from pharmpy.tools.mfl.statement.feature.symbols import Wildcard
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.tools.scm.results import candidate_summary_dataframe, ofv_summary_dataframe
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

from ..mfl.filter import COVSEARCH_STATEMENT_TYPES
from ..mfl.parse import ModelFeatures, get_model_features
from .results import COVSearchResults

NAME_WF = 'covsearch'

DataFrame = Any  # NOTE: should be pd.DataFrame but we want lazy loading


@dataclass(frozen=True)
class Effect:
    parameter: str
    covariate: str
    fp: str
    operation: str


class AddEffect(Effect):
    pass


class RemoveEffect(Effect):
    pass


@dataclass(frozen=True)
class Step:
    alpha: float
    effect: Effect


class ForwardStep(Step):
    pass


class BackwardStep(Step):
    pass


def _added_effects(steps: Tuple[Step, ...]) -> Iterable[Effect]:
    added_effects = defaultdict(list)
    for i, step in enumerate(steps):
        if isinstance(step, ForwardStep):
            added_effects[astuple(step.effect)].append(i)
        else:
            assert isinstance(step, BackwardStep)
            added_effects[astuple(step.effect)].pop()

    pos = {effect: set(indices) for effect, indices in added_effects.items()}

    for i, step in enumerate(steps):
        if isinstance(step, ForwardStep) and i in pos[astuple(step.effect)]:
            yield step.effect


@dataclass
class Candidate:
    modelentry: ModelEntry
    steps: Tuple[Step, ...]


@dataclass
class SearchState:
    user_input_modelentry: ModelEntry
    start_modelentry: ModelEntry
    best_candidate_so_far: Candidate
    all_candidates_so_far: List[Candidate]


ALGORITHMS = ('scm-forward', 'scm-forward-then-backward')


def create_workflow(
    effects: Union[str, Sequence[InputSpec]],
    p_forward: float = 0.01,
    p_backward: float = 0.001,
    max_steps: int = -1,
    algorithm: Literal[ALGORITHMS] = 'scm-forward-then-backward',
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
    naming_index_offset: Optional[int] = 0,
):
    """Run COVsearch tool. For more details, see :ref:`covsearch`.

    Parameters
    ----------
    effects : str
        MFL of covariate effects to try
    p_forward : float
        The p-value to use in the likelihood ratio test for forward steps
    p_backward : float
        The p-value to use in the likelihood ratio test for backward steps
    max_steps : int
        The maximum number of search steps to make
    algorithm : {'scm-forward', 'scm-forward-then-backward'}
        The search algorithm to use. Currently, 'scm-forward' and
        'scm-forward-then-backward' are supported.
    results : ModelfitResults
        Results of model
    model : Model
        Pharmpy model
    strictness : str or None
        Strictness criteria
    naming_index_offset: int
        index offset for naming of runs. Default is 0.

    Returns
    -------
    COVSearchResults
        COVsearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import run_covsearch, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> effects = 'COVARIATE([CL, V], [AGE, WT], EXP)'
    >>> res = run_covsearch(effects, model=model, results=results)      # doctest: +SKIP
    """

    wb = WorkflowBuilder(name=NAME_WF)

    # FIXME : Handle when model is None
    start_task = Task("create_modelentry", _start, model, results)
    init_task = Task("init", _init_search_state, effects)
    wb.add_task(init_task, predecessors=start_task)

    forward_search_task = Task(
        'forward-search',
        task_greedy_forward_search,
        p_forward,
        max_steps,
        naming_index_offset,
        strictness,
    )

    wb.add_task(forward_search_task, predecessors=init_task)
    search_output = wb.output_tasks

    if algorithm == 'scm-forward-then-backward':
        backward_search_task = Task(
            'backward-search',
            task_greedy_backward_search,
            p_backward,
            max_steps,
            naming_index_offset,
            strictness,
        )

        wb.add_task(backward_search_task, predecessors=search_output)
        search_output = wb.output_tasks

    results_task = Task(
        'results',
        task_results,
        p_forward,
        p_backward,
        strictness,
    )

    wb.add_task(results_task, predecessors=search_output)

    return Workflow(wb)


def _start(model, results):
    return ModelEntry.create(model=model, parent=None, modelfit_results=results)


def _init_search_state(context, effects: str, modelentry: ModelEntry) -> SearchState:
    model = modelentry.model
    effect_funcs, filtered_model = filter_search_space_and_model(effects, model)
    if filtered_model != model:
        filtered_modelentry = ModelEntry.create(model=filtered_model)
        filtered_fit_wf = create_fit_workflow(modelentries=[filtered_modelentry])
        filtered_modelentry = call_workflow(filtered_fit_wf, 'fit_filtered_model', context)
    else:
        filtered_modelentry = modelentry
    candidate = Candidate(filtered_modelentry, ())
    return (effect_funcs, SearchState(modelentry, filtered_modelentry, candidate, [candidate]))


def filter_search_space_and_model(effects, model):
    filtered_model = model.replace(name="filtered_input_model", parent_model="filtered_input_model")
    ss_mfl = ModelFeatures.create_from_mfl_string(effects).expand(
        filtered_model
    )  # Expand to remove LET/REF
    model_mfl = ModelFeatures.create_from_mfl_string(get_model_features(filtered_model))

    # Remove all covariates not part of the search space
    to_be_removed = model_mfl - ss_mfl
    covariate_list = to_be_removed.mfl_statement_list(["covariate"])
    description = []
    if len(covariate_list) != 0:
        description.append("REMOVED")
        for cov_effect in parse_spec(spec(filtered_model, covariate_list)):
            if cov_effect[2].lower() == "custom":
                try:
                    filtered_model = remove_covariate_effect(
                        filtered_model, cov_effect[0], cov_effect[1]
                    )
                    description.append(f'({cov_effect[0]}-{cov_effect[1]}-{cov_effect[2]})')

                except AssertionError:
                    # All custom effects are grouped and therefor might not exist
                    # FIXME : remove_covariate_effect not raise if non-existing ?
                    pass
            else:
                filtered_model = remove_covariate_effect(
                    filtered_model, cov_effect[0], cov_effect[1]
                )
                description.append(f'({cov_effect[0]}-{cov_effect[1]}-{cov_effect[2]})')
        filtered_model = filtered_model.replace(description=';'.join(description))

    ss_funcs = ss_mfl.convert_to_funcs(["covariate"])

    remove_funcs = {k: v for k, v in ss_funcs.items() if k[-1] == "REMOVE"}
    optional_funcs = {}
    for key in remove_funcs.keys():
        add_key = key[:-1] + ("ADD",)
        optional_funcs[add_key] = ss_funcs[add_key]
    mandatory_funcs = {
        k: v
        for k, v in ss_funcs.items()
        if k[-1] != "REMOVE" and k[:-1] + ("REMOVE",) not in ss_funcs.keys()
    }

    def func_description(effect_funcs, model=None, add=True):
        d = []
        for eff_descriptor, _ in effect_funcs.items():
            if model:
                if (
                    has_covariate_effect(model, eff_descriptor[1], eff_descriptor[2])
                    if add
                    else not has_covariate_effect(model, eff_descriptor[1], eff_descriptor[2])
                ):
                    d.append(f'({eff_descriptor[1]}-{eff_descriptor[2]}-{eff_descriptor[3]})')
            else:
                d.append(f'({eff_descriptor[1]}-{eff_descriptor[2]}-{eff_descriptor[3]})')
        return d

    # Remove all optional covariates
    if len(remove_funcs) != 0:
        potential_extension = func_description(remove_funcs, filtered_model, add=True)
        for _, optional_func in remove_funcs.items():
            filtered_model = optional_func(filtered_model)
        if len(potential_extension) != 0:
            if not description:
                description.append("REMOVED")
            description.extend(potential_extension)

    # Add all mandatory covariates
    if len(mandatory_funcs) != 0:
        change_description = True
        for eff_descriptor, mandatory_func in mandatory_funcs.items():
            if not has_covariate_effect(filtered_model, eff_descriptor[1], eff_descriptor[2]):
                if change_description:
                    description.append("ADDED")
                    description.extend(func_description(mandatory_funcs, filtered_model, add=False))
                    change_description = False
                filtered_model = mandatory_func(filtered_model)

    # Filter unneccessary keys from fuctions
    optional_funcs = {k[1:-1]: v for k, v in optional_funcs.items()}

    if len(description) > 1:
        filtered_model = filtered_model.replace(description=';'.join(description))
        return (optional_funcs, filtered_model)
    else:
        return (optional_funcs, model)


def task_greedy_forward_search(
    context,
    p_forward: float,
    max_steps: int,
    naming_index_offset: int,
    strictness: Optional[str],
    state_and_effect: Tuple[SearchState, dict],
) -> SearchState:
    for temp in state_and_effect:
        if isinstance(temp, SearchState):
            state = temp
        else:
            candidate_effect_funcs = temp
    candidate = state.best_candidate_so_far
    assert state.all_candidates_so_far == [candidate]

    def handle_effects(
        step: int,
        parent: Candidate,
        candidate_effect_funcs: dict,
        index_offset: int,
    ):
        index_offset = index_offset + naming_index_offset
        wf = wf_effects_addition(parent.modelentry, parent, candidate_effect_funcs, index_offset)
        new_candidate_modelentries = call_workflow(
            wf, f'{NAME_WF}-effects_addition-{step}', context
        )
        return [
            Candidate(modelentry, parent.steps + (ForwardStep(p_forward, AddEffect(*effect)),))
            for modelentry, effect in zip(new_candidate_modelentries, candidate_effect_funcs.keys())
        ]

    return _greedy_search(
        state,
        handle_effects,
        candidate_effect_funcs,
        p_forward,
        max_steps,
        strictness,
    )


def task_greedy_backward_search(
    context,
    p_backward: float,
    max_steps: int,
    naming_index_offset,
    strictness: Optional[str],
    state: SearchState,
) -> SearchState:
    def handle_effects(
        step: int,
        parent: Candidate,
        candidate_effect_funcs: List[EffectLiteral],
        index_offset: int,
    ):
        index_offset = index_offset + naming_index_offset
        wf = wf_effects_removal(parent, candidate_effect_funcs, index_offset)
        new_candidate_modelentries = call_workflow(wf, f'{NAME_WF}-effects_removal-{step}', context)

        return [
            Candidate(modelentry, parent.steps + (BackwardStep(p_backward, RemoveEffect(*effect)),))
            for modelentry, effect in zip(new_candidate_modelentries, candidate_effect_funcs.keys())
        ]

    optional_effects = list(map(astuple, _added_effects(state.best_candidate_so_far.steps)))
    candidate_effect_funcs = dict(
        covariate_features(
            state.best_candidate_so_far.modelentry.model,
            ModelFeatures.create_from_mfl_string(
                get_model_features(state.best_candidate_so_far.modelentry.model)
            ).covariate,
            remove=True,
        )
    )

    candidate_effect_funcs = {
        k[1:-1]: v for k, v in candidate_effect_funcs.items() if (*k[1:-1],) in optional_effects
    }

    n_removable_effects = max(0, len(state.best_candidate_so_far.steps) - 1)

    return _greedy_search(
        state,
        handle_effects,
        candidate_effect_funcs,
        p_backward,
        min(max_steps, n_removable_effects) if max_steps >= 0 else n_removable_effects,
        strictness,
    )


def _greedy_search(
    state: SearchState,
    handle_effects: Callable[[int, Candidate, List[EffectLiteral], int], List[Candidate]],
    candidate_effect_funcs: dict,
    alpha: float,
    max_steps: int,
    strictness: Optional[str],
) -> SearchState:
    best_candidate_so_far = state.best_candidate_so_far
    all_candidates_so_far = list(
        state.all_candidates_so_far
    )  # NOTE: This includes start model/filtered model

    steps = range(1, max_steps + 1) if max_steps >= 0 else count(1)

    for step in steps:
        if not candidate_effect_funcs:
            break

        new_candidates = handle_effects(
            step, best_candidate_so_far, candidate_effect_funcs, len(all_candidates_so_far) - 1
        )

        all_candidates_so_far.extend(new_candidates)
        new_candidate_modelentries = list(
            map(lambda candidate: candidate.modelentry, new_candidates)
        )

        parent_modelentry = best_candidate_so_far.modelentry
        ofvs = [
            np.nan
            if modelentry.modelfit_results is None
            or not is_strictness_fulfilled(
                modelentry.modelfit_results, modelentry.model, strictness
            )
            else modelentry.modelfit_results.ofv
            for modelentry in new_candidate_modelentries
        ]
        # NOTE: We assume parent_modelentry.modelfit_results is not None
        assert parent_modelentry.modelfit_results is not None
        best_model_so_far = lrt_best_of_many(
            parent_modelentry,
            new_candidate_modelentries,
            parent_modelentry.modelfit_results.ofv,
            ofvs,
            alpha,
        )

        if best_model_so_far is parent_modelentry:
            break

        best_candidate_so_far = next(
            filter(
                lambda candidate: candidate.modelentry is best_model_so_far, all_candidates_so_far
            )
        )

        # NOTE: Filter out incompatible effects
        last_step_effect = best_candidate_so_far.steps[-1].effect

        candidate_effect_funcs = {
            effect_description: effect_func
            for effect_description, effect_func in candidate_effect_funcs.items()
            if effect_description[0] != last_step_effect.parameter
            or effect_description[1] != last_step_effect.covariate
        }

    return SearchState(
        state.user_input_modelentry,
        state.start_modelentry,
        best_candidate_so_far,
        all_candidates_so_far,
    )


def wf_effects_addition(
    modelentry: ModelEntry,
    candidate: Candidate,
    candidate_effect_funcs: dict,
    index_offset: int,
):
    wb = WorkflowBuilder()

    for i, effect in enumerate(candidate_effect_funcs.items(), 1):
        task = Task(
            repr(effect[0]),
            task_add_covariate_effect,
            modelentry,
            candidate,
            effect,
            index_offset + i,
        )
        wb.add_task(task)

    wf_fit = create_fit_workflow(n=len(candidate_effect_funcs))
    wb.insert_workflow(wf_fit)

    task_gather = Task('gather', lambda *models: models)
    wb.add_task(task_gather, predecessors=wb.output_tasks)
    return Workflow(wb)


def task_add_covariate_effect(
    modelentry: ModelEntry, candidate: Candidate, effect: dict, effect_index: int
):
    model = modelentry.model
    name = f'covsearch_run{effect_index}'
    description = _create_description(effect[0], candidate.steps)
    model_with_added_effect = model.replace(name=name, description=description)
    model_with_added_effect = update_initial_estimates(
        model_with_added_effect, modelentry.modelfit_results
    )
    func = effect[1]
    model_with_added_effect = func(model_with_added_effect, allow_nested=True)
    return ModelEntry.create(
        model=model_with_added_effect, parent=candidate.modelentry.model, modelfit_results=None
    )


def _create_description(effect_new: dict, steps_prev: Tuple[Step, ...], forward: bool = True):
    # Will create this type of description: '(CL-AGE-exp);(MAT-AGE-exp);(MAT-AGE-exp-+)'
    def _create_effect_str(effect):
        if isinstance(effect, Tuple):
            param, cov, fp, op = effect
        elif isinstance(effect, Effect):
            param, cov, fp, op = effect.parameter, effect.covariate, effect.fp, effect.operation
        else:
            raise ValueError('Effect must be a tuple or Effect dataclass')
        effect_base = f'{param}-{cov}-{fp}'
        if op == '+':
            effect_base += f'-{op}'
        return f'({effect_base})'

    effect_new_str = _create_effect_str(effect_new)
    effects = []
    for effect_prev in _added_effects(steps_prev):
        effect_prev_str = _create_effect_str(effect_prev)
        if not forward and effect_prev_str == effect_new_str:
            continue
        effects.append(effect_prev_str)
    if forward:
        effects.append(effect_new_str)
    return ';'.join(effects)


def wf_effects_removal(
    parent: Candidate,
    candidate_effect_funcs: dict,
    index_offset: int,
):
    wb = WorkflowBuilder()

    for i, effect in enumerate(candidate_effect_funcs.items(), 1):
        task = Task(
            repr(effect[0]),
            task_remove_covariate_effect,
            parent,
            effect,
            index_offset + i,
        )
        wb.add_task(task)

    wf_fit = create_fit_workflow(n=len(candidate_effect_funcs))
    wb.insert_workflow(wf_fit)

    task_gather = Task('gather', lambda *models: models)
    wb.add_task(task_gather, predecessors=wb.output_tasks)
    return Workflow(wb)


def task_remove_covariate_effect(candidate: Candidate, effect: dict, effect_index: int):
    model = candidate.modelentry.model
    name = f'covsearch_run{effect_index}'
    description = _create_description(effect[0], candidate.steps, forward=False)
    model_with_removed_effect = model.replace(
        name=name, description=description, parent_model=model.name
    )

    func = effect[1]
    model_with_removed_effect = func(model_with_removed_effect)

    model_with_removed_effect = update_initial_estimates(model_with_removed_effect)
    return ModelEntry.create(
        model=model_with_removed_effect, parent=candidate.modelentry.model, modelfit_results=None
    )


def task_results(p_forward: float, p_backward: float, strictness: str, state: SearchState):
    candidates = state.all_candidates_so_far
    modelentries = list(map(lambda candidate: candidate.modelentry, candidates))
    base_modelentry, *res_modelentries = modelentries
    assert base_modelentry is state.start_modelentry
    best_modelentry = state.best_candidate_so_far.modelentry
    user_input_modelentry = state.user_input_modelentry
    if user_input_modelentry != base_modelentry:
        modelentries = [user_input_modelentry] + modelentries

    res = create_results(
        COVSearchResults,
        base_modelentry,
        base_modelentry,
        res_modelentries,
        'lrt',
        (p_forward, p_backward),
    )

    steps = _make_df_steps(best_modelentry, candidates)
    res = replace(
        res,
        final_model=best_modelentry.model,
        steps=steps,
        candidate_summary=candidate_summary_dataframe(steps),
        ofv_summary=ofv_summary_dataframe(steps, final_included=True, iterations=True),
        summary_tool=_modify_summary_tool(res.summary_tool, steps),
        summary_models=_summarize_models(modelentries, steps),
    )

    return res


def _modify_summary_tool(summary_tool, steps):
    step_cols_to_keep = ['step', 'pvalue', 'goal_pvalue', 'is_backward', 'selected', 'model']
    steps_df = steps.reset_index()[step_cols_to_keep].set_index(['step', 'model'])

    summary_tool_new = steps_df.join(summary_tool)
    column_to_move = summary_tool_new.pop('description')

    summary_tool_new.insert(0, 'description', column_to_move)
    return summary_tool_new.drop(['rank'], axis=1)


def _summarize_models(modelentries, steps):
    summary_models = summarize_modelfit_results(
        [modelentry.modelfit_results for modelentry in modelentries]
    )
    summary_models['step'] = steps.reset_index().set_index(['model'])['step']

    return summary_models.reset_index().set_index(['step', 'model'])


def _make_df_steps(best_modelentry: ModelEntry, candidates: List[Candidate]):
    best_model = best_modelentry.model
    modelentries_dict = {
        candidate.modelentry.model.name: candidate.modelentry for candidate in candidates
    }
    children_count = Counter(
        candidate.modelentry.parent.name for candidate in candidates if candidate.modelentry.parent
    )

    # Find if longest forward is also the input for the backwards search
    # otherwise add an index offset to _make_df_steps_function
    forward_candidates = [
        fc for fc in candidates if fc.steps == tuple() or isinstance(fc.steps[-1], ForwardStep)
    ]
    largest_forward_step = max([len(fc.steps) for fc in forward_candidates])
    largest_forward_candidates = [
        fc for fc in forward_candidates if len(fc.steps) == largest_forward_step
    ]
    index_offset = 0
    if not any(
        children_count[c.modelentry.model.name] >= 1 or c.modelentry.model is best_model
        for c in largest_forward_candidates
    ):
        index_offset = 1

    data = (
        _make_df_steps_row(
            modelentries_dict, children_count, best_model, candidate, index_offset=index_offset
        )
        for candidate in candidates
    )

    return pd.DataFrame.from_records(
        data,
        index=['step', 'parameter', 'covariate', 'extended_state'],
    )


def _make_df_steps_row(
    modelentries_dict: dict,
    children_count: Counter,
    best_model: Model,
    candidate: Candidate,
    index_offset=0,
):
    modelentry = candidate.modelentry
    model = modelentry.model
    parent_name = modelentry.parent.name if modelentry.parent else model.name
    parent_modelentry = modelentries_dict[parent_name]
    parent_model = parent_modelentry.model
    if candidate.steps:
        last_step = candidate.steps[-1]
        last_effect = last_step.effect
        parameter, covariate = last_effect.parameter, last_effect.covariate
        extended_state = f'{last_effect.operation} {last_effect.fp}'
        is_backward = isinstance(last_step, BackwardStep)
        if not is_backward:
            reduced_ofv = np.nan if (mfr := parent_modelentry.modelfit_results) is None else mfr.ofv
            extended_ofv = np.nan if (mfr := modelentry.modelfit_results) is None else mfr.ofv
            alpha = last_step.alpha
            extended_significant = lrt_test(
                parent_model,
                model,
                reduced_ofv,
                extended_ofv,
                alpha,
            )
        else:
            extended_ofv = (
                np.nan if (mfr := parent_modelentry.modelfit_results) is None else mfr.ofv
            )
            reduced_ofv = np.nan if (mfr := modelentry.modelfit_results) is None else mfr.ofv
            alpha = last_step.alpha
            extended_significant = lrt_test(
                model,
                parent_model,
                reduced_ofv,
                extended_ofv,
                alpha,
            )
        ofv_drop = reduced_ofv - extended_ofv
    else:
        parameter, covariate, extended_state = '', '', ''
        is_backward = False
        reduced_ofv = np.nan if (mfr := parent_modelentry.modelfit_results) is None else mfr.ofv
        extended_ofv = np.nan if (mfr := modelentry.modelfit_results) is None else mfr.ofv
        ofv_drop = reduced_ofv - extended_ofv
        alpha, extended_significant = np.nan, np.nan

    selected = children_count[model.name] >= 1 or model.name == best_model.name
    if not is_backward:
        p_value = lrt_p_value(parent_model, model, reduced_ofv, extended_ofv)
        assert not selected or (model is parent_model) or extended_significant
    else:
        p_value = lrt_p_value(model, parent_model, reduced_ofv, extended_ofv)
        assert not selected or (model is parent_model) or not extended_significant

    return {
        'step': len(candidate.steps)
        if candidate.steps == tuple() or isinstance(candidate.steps[-1], ForwardStep)
        else len(candidate.steps) + index_offset,
        'parameter': parameter,
        'covariate': covariate,
        'extended_state': extended_state,
        'reduced_ofv': reduced_ofv,
        'extended_ofv': extended_ofv,
        'ofv_drop': ofv_drop,
        'delta_df': len(model.parameters.nonfixed) - len(parent_model.parameters.nonfixed),
        'pvalue': p_value,
        'goal_pvalue': alpha,
        'is_backward': is_backward,
        'extended_significant': extended_significant,
        'selected': selected,
        'model': model.name,
        'covariate_effects': np.nan,
    }


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    effects, p_forward, p_backward, algorithm, model, strictness, naming_index_offset
):
    if not 0 < p_forward <= 1:
        raise ValueError(
            f'Invalid `p_forward`: got `{p_forward}`, must be a float in range (0, 1].'
        )

    if not 0 < p_backward <= 1:
        raise ValueError(
            f'Invalid `p_backward`: got `{p_backward}`, must be a float in range (0, 1].'
        )

    if model is not None:
        try:
            statements = mfl_parse(effects)
        except:  # noqa E722
            raise ValueError(f'Invalid `effects`, could not be parsed: `{effects}`')

        bad_statements = list(
            filter(
                lambda statement: not isinstance(statement, COVSEARCH_STATEMENT_TYPES),
                statements,
            )
        )

        if bad_statements:
            raise ValueError(
                f'Invalid `effects`: found unknown statement of type {type(bad_statements[0]).__name__}.'
            )

        for s in statements:
            if isinstance(s, Covariate) and isinstance(s.fp, Wildcard) and not s.optional.option:
                raise ValueError(
                    f'Invalid `effects` due to non-optional covariate'
                    f' defined with WILDCARD as effect in {s}'
                    f' Only single effect allowed for mandatory covariates'
                )

        effect_spec = spec(model, statements)

        candidate_effects = map(
            lambda x: Effect(*x[:-1]), sorted(set(parse_spec(effect_spec)))
        )  # Ignore OPTIONAL attribute

        allowed_covariates = get_covariates_allowed_in_covariate_effect(model)
        allowed_parameters = set(get_pk_parameters(model)).union(
            str(statement.symbol) for statement in model.statements.before_odes
        )
        allowed_covariate_effects = set(all_covariate_effects)
        allowed_ops = set(['*', '+'])

        for effect in candidate_effects:
            if effect.covariate not in allowed_covariates:
                raise ValueError(
                    f'Invalid `effects` because of invalid covariate found in'
                    f' effects: got `{effect.covariate}`,'
                    f' must be in {sorted(allowed_covariates)}.'
                )
            if effect.parameter not in allowed_parameters:
                raise ValueError(
                    f'Invalid `effects` because of invalid parameter found in'
                    f' effects: got `{effect.parameter}`,'
                    f' must be in {sorted(allowed_parameters)}.'
                )
            if effect.fp not in allowed_covariate_effects:
                raise ValueError(
                    f'Invalid `effects` because of invalid effect function found in'
                    f' effects: got `{effect.fp}`,'
                    f' must be in {sorted(allowed_covariate_effects)}.'
                )
            if effect.operation not in allowed_ops:
                raise ValueError(
                    f'Invalid `effects` because of invalid effect operation found in'
                    f' effects: got `{effect.operation}`,'
                    f' must be in {sorted(allowed_ops)}.'
                )
    if strictness is not None and "rse" in strictness.lower():
        if model.estimation_steps[-1].parameter_uncertainty_method is None:
            raise ValueError(
                'parameter_uncertainty_method not set for model, cannot calculate relative standard errors.'
            )
    if not isinstance(naming_index_offset, int) or naming_index_offset < 0:
        raise ValueError('naming_index_offset need to be a postive (>=0) integer.')



from pathlib import Path
from typing import Optional, Union

from pharmpy.internals.fs.path import normalize_user_given_path, path_absolute
from pharmpy.model import Model


def create_dataset_path(model: Model, path: Optional[Union[str, Path]] = None) -> Path:
    path = path_absolute(Path("" if path is None else path))

    if path and not path.is_dir():
        return path

    if (di_path := model.datainfo.path) is None:
        filename = f"{model.name}.csv"
    else:
        filename = di_path.with_suffix('.csv').name

    return path / filename


def write_csv(model: Model, path: Optional[Union[str, Path]] = None, force: bool = False) -> Model:
    """Write dataset to a csv file and updates the datainfo path

    Parameters
    ----------
    model : Model
        Model whose dataset to write to file
    path : None or str or Path
        Destination path. Default is to use original path with .csv suffix.
    force : bool
        Overwrite file with same path. Default is False.

    Returns
    -------
    Model
       Updated model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, write_csv
    >>> model = load_example_model("pheno")
    >>> model = write_csv(model, path="newdataset.csv")    # doctest: +SKIP

    """
    from pharmpy.model import data

    if path is not None:
        path = normalize_user_given_path(path)

    path = create_dataset_path(model, path)
    if not force and path.exists():
        raise FileExistsError(f'File at {path} already exists.')

    path = path_absolute(path)
    model.dataset.to_csv(path, na_rep=data.conf.na_rep, index=False)
    model = model.replace(datainfo=model.datainfo.replace(path=path))
    return model



import pytest

import pharmpy.model.external.nonmem.nmtran_parser
import pharmpy.model.external.nonmem.records


@pytest.fixture
def records():
    return pharmpy.model.external.nonmem.records


@pytest.fixture
def parser():
    return pharmpy.model.external.nonmem.nmtran_parser.NMTranParser()



from functools import wraps
from inspect import Signature, signature
from typing import Callable


def with_same_arguments_as(ref: Callable):
    ref_signature = signature(ref)
    ref_args, ref_defaults = _signature_map(ref_signature)
    ref_args_set = set(ref_args)
    ref_index = {arg: i for i, arg in enumerate(ref_args)}

    def _lookup(args, kwargs, key):
        i = ref_index[key]
        if i < len(args):
            return args[i]
        try:
            return kwargs[key]
        except KeyError:
            return ref_defaults[key]

    def _with_same_arguments(fn: Callable):
        fn_args, fn_defaults = _signature_map(signature(fn))

        assert set(fn_args) <= ref_args_set
        assert not fn_defaults

        @wraps(fn)
        def _wrapped(*args, **kwargs):
            if len(args) > len(ref_args):
                raise TypeError(
                    f'{fn.__name__}() takes {len(ref_args)} but {len(args)} where given'
                )

            for arg in kwargs:
                if arg not in ref_args_set:
                    raise TypeError(
                        f'{fn.__name__}() got an unexpected keyword argument: \'{arg}\''
                    )

            try:
                permuted = [_lookup(args, kwargs, arg) for arg in fn_args]
            except KeyError as e:
                arg = e.args[0]
                raise TypeError(
                    f'{fn.__name__}() missing 1 required positional argument: \'{arg}\''
                )

            return fn(*permuted)

        # NOTE: Lazy annotations can only be resolved properly if the
        # declarations of `ref` and `fn` have the same locals() and
        # globals().
        _wrapped.__annotations__ = ref.__annotations__
        _wrapped.__signature__ = ref_signature  # pyright: ignore [reportGeneralTypeIssues]

        return _wrapped

    return _with_same_arguments


def _signature_map(ref_signature: Signature):
    ref_args = []
    ref_defaults = {}
    for param in ref_signature.parameters.values():
        assert param.kind != param.VAR_POSITIONAL  # TODO: Handle varargs
        assert param.kind != param.VAR_KEYWORD  # TODO: Handle kwargs
        name = param.name
        ref_args.append(name)
        default = param.default
        if default is not param.empty:
            ref_defaults[name] = default

    return ref_args, ref_defaults
